/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.common;

import java.io.File;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Properties;
import org.objectweb.common.TraceCore;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_jms.TraceJms;
import org.objectweb.jonas_timer.TraceTimer;
import org.objectweb.jonas_tm.TraceTm;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.wrapper.printwriter.PrintWriterImpl;
import org.objectweb.util.monolog.wrapper.p6spy.P6SpyLogger;
import org.objectweb.util.monolog.file.DottedStringTools;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;

/**
 *  This class provides utility method for using Monolog
 * @author Philippe Coq.
 * Sebastien Chassande-Barrioz sebastien.chassande@inrialpes.fr
 * Philippe Durieux    
 * Florent Benoit & Ludovic Bert  
 * Adriana Danes:
 * 2003/04/08 : change MBean implementation model from inheritance to delegation   
 */
public class Log {    

    // name of the config file - initialized by the configure method
    private static String configFile = null;

    /**
     * This class is actually a wrapper on the unique loggerFactory
     */
    private static LoggerFactory lf = null;

    /**
     * Properties used for Logging system
     * Kept here mainly for debugging
     */
    private static Properties props = null;

    // List of topic used inside JOnAS
    // This list may be not exhaustive since many topics can be added for
    // different purposes or to refine debugging.
    // To get the complete list, use "jonas admin -t"
    public static final String JONAS_DBM_PREFIX = "org.objectweb.jonas.dbm";
    public static final String JONAS_JCA_PREFIX = "org.objectweb.jonas.jca";
    public static final String JONAS_JDBCXA_PREFIX = "org.objectweb.jonas.jdbc_xa";
    public static final String JONAS_NAMING_PREFIX = "org.objectweb.jonas.naming";
    public static final String JONAS_PROPCTX_PREFIX = "org.objectweb.jonas.propagation";
    public static final String JONAS_REGISTRY_PREFIX = "org.objectweb.jonas.registry";
    public static final String JONAS_SECURITY_PREFIX = "org.objectweb.jonas.security";
    public static final String JONAS_SERVER_PREFIX = "org.objectweb.jonas.server";
    public static final String JONAS_WEB_PREFIX = "org.objectweb.jonas.web";
    public static final String JONAS_EAR_PREFIX = "org.objectweb.jonas.ear";
    public static final String JONAS_MAIL_PREFIX = "org.objectweb.jonas.mail";
    public static final String JONAS_TESTS_PREFIX = "org.objectweb.jonas_tests";
    public static final String JONAS_LOADER_PREFIX = "org.objectweb.jonas.loader";
    public static final String JONAS_GENIC_PREFIX = "org.objectweb.jonas.genic";
    public static final String JONAS_MANAGEMENT_PREFIX = "org.objectweb.jonas.management";
    public static final String SPY_LOGGER_NAME = "org.objectweb.jonas.jdbc.sql";

    /**
     * Output tag for the hamdler
     */
    private static final String HANDLER_OUTPUT_ATTRIBUTE = "output";

    /**
     * Output tag which must be replaced by JOnAS
     */
    private static final String AUTOMATIC_CONFIG = "automatic";

    /**
     * Directory for the logs $JONAS_BASE/logs
     */
    private static final String LOG_DIRECTORY = "logs";

    /**
     * Format of a timestamp log.
     */
    private static final String TIMESTAMP_FORMAT = "-yyyy-MM-dd";

    /**
     * Suffix of a log file.
     */
    private static final String SUFFIX_LOGFILE = "log";

    /**
     * Configure Logger
     * @param file The configuration file for monolog (usually: trace.properties)
     */
    public static void configure(String file) {
        configFile = file;
        getLoggerFactory();
        P6SpyLogger.logger = lf.getLogger(SPY_LOGGER_NAME);
        TraceTimer.configure(lf);
        TraceCore.configure(lf);
        TraceTm.configure(lf);
        TraceEjb.configure(lf);
        TraceJms.configure(lf);
    }

    /**
     * It returns the unique LoggerFactory used in JOnAS.
     * initialize it if not already done.
     */
    public static synchronized LoggerFactory getLoggerFactory() {
        if (lf == null) {

            try {

                // Detect client case
                // get jonas configuration properties
                Object jProp = null;
                try {
                    jProp = JProp.getInstance();
                    // Run jonas 
                    // Create a JProp object for the configFile
                    JProp p = JProp.getInstance(configFile);
                } catch (java.lang.NoClassDefFoundError ncdfe) {
                    // Run in a jclient.
                    // Define a simple monolog
                    props = new Properties();
                    props.put("log.config.classname", "org.objectweb.util.monolog.wrapper.log4j.MonologLoggerFactory");
                    // Debug property ?
                    if (System.getProperty("DEBUG") != null) {
                        props.put("logger.root.level", "DEBUG");
                    } else {
                        props.put("logger.root.level", "ERROR");
                    }
                    return Monolog.init(props);
                } catch (Exception e) {
                    System.out.println("Can't read jonas.properties. Check that you have defined a $JONAS_BASE variable");
                }

                


                // load properties from config file
                props = JProp.getInstance(configFile).getConfigFileEnv();
                String jonasBase = null;
                String nameOfServer = null;
                if (jProp != null) {
                    jonasBase = ((JProp) jProp).getJonasBase();
                    nameOfServer = ((JProp) jProp).getValue("jonas.name", "jonas");
                } else {
                    // Set default
                    jonasBase = ".";
                    nameOfServer = "jonas";
                }

                // Before sending props, replace AUTOMATIC pattern by a file in $JONAS_BASE/logs/
                // File are named <JOnAS_server_name>TIMESTAMP_FORMAT.SUFFIX_LOGFILE
                File logFirFile = new File(jonasBase + File.separator + LOG_DIRECTORY);

                // Properties sent may be updated if AUTOMATIC_CONFIG is used
                // Updated properties are in
                Properties updatedProps = null;

                //Create a date format
                SimpleDateFormat sdf = new SimpleDateFormat(TIMESTAMP_FORMAT);
                String date = sdf.format(new Date());
                // How many AUTOMATIC field have been replaced
                int nbAut = 0;

                File logFile = null;
                for (Enumeration enum = props.keys(); enum.hasMoreElements();) {
                    String key = (String) enum.nextElement();
                    if (key == null) {
                        continue;
                    }
                    
                    if (key.startsWith(PropertiesConfAccess.HANDLER_FIELD)) {
                        String temp = DottedStringTools.getFirst(key);

                        if (temp == null) {
                            continue;
                        }

                        temp = DottedStringTools.getEnd(key);
                        if (temp == null) {
                            continue;
                        }

                        String handlerName = DottedStringTools.getBegin(temp);

                        String stringType = props.getProperty(PropertiesConfAccess.HANDLER_FIELD + PropertiesConfAccess.DOT + handlerName + PropertiesConfAccess.DOT + PropertiesConfAccess.HANDLER_TYPE_ATTRIBUTE, null);
                        String stringOutput = props.getProperty(PropertiesConfAccess.HANDLER_FIELD + PropertiesConfAccess.DOT + handlerName + PropertiesConfAccess.DOT + HANDLER_OUTPUT_ATTRIBUTE, null);
                        // Got the type and the output
                        // --> Compare with the pattern AUTOMATIC_CONFIG for the FILE type.
                        if ((stringType != null) && (stringOutput != null)) {
                            if (PropertiesConfAccess.HANDLER_TYPE_ATTRIBUTE_FILE_VALUE.equalsIgnoreCase(stringType)
                                || (PropertiesConfAccess.HANDLER_TYPE_ATTRIBUTE_ROLLING_FILE_VALUE.equalsIgnoreCase(stringType))) {
                                if (stringOutput.equalsIgnoreCase(AUTOMATIC_CONFIG)) {
                                    // OK, it's match
                                    String fileName = null;

                                    if (nbAut > 0) {
                                        // Add an unique entry by adding nbAut value.
                                        fileName = nameOfServer + date + PropertiesConfAccess.DOT + nbAut + PropertiesConfAccess.DOT + SUFFIX_LOGFILE;
                                    } else {
                                        // this is the first occurence of AUTOMATIC_CONFIG, create updatedProps
                                        updatedProps = (Properties) props.clone();
                                        fileName = nameOfServer + date + PropertiesConfAccess.DOT + SUFFIX_LOGFILE;
                                    }

                                    // Log directory exists ?
                                    if (logFirFile.exists()) {
                                        logFile = new File(logFirFile, fileName);
                                    } else {
                                        // No $JONAS_BASE/logs directory, try current directory
                                        logFile = new File(fileName);
                                    }

                                    updatedProps.setProperty(PropertiesConfAccess.HANDLER_FIELD + PropertiesConfAccess.DOT + handlerName + PropertiesConfAccess.DOT + HANDLER_OUTPUT_ATTRIBUTE, logFile.getPath());
                                    nbAut++;

                                }
                            }
                        }
                    }
                }

                // Instanciate the LoggerFactory
                String b = props.getProperty("log.config.classname", null);
                if (b == null) {
                    System.err.println("Malformed configuration log file: log.config.classname not available");
                    return null;
                }
                lf = (LoggerFactory) Class.forName(b).newInstance();
                // Configure the LoggerFactory with the updated properties or the initial ones
                if (updatedProps == null) {
                    PropertiesConfAccess.load(props, lf, (HandlerFactory) lf, (LevelFactory) lf);
                } else {
                    PropertiesConfAccess.load(updatedProps, lf, (HandlerFactory) lf, (LevelFactory) lf);
                }
            } catch (Exception e) {
                System.err.println("Cannot get LoggerFactory:" + e);
                e.printStackTrace();
            }
        }
        return lf;
    }

    /**
     * Returns the standard PrintWriter associated to the logger defined by
     * its topic.
     * This is mainly used for DBM and Connectors.
     */
    public static PrintWriter getLogWriter(String topic) {
        // TODO : should not create a new object at each call
        return new PrintWriterImpl(getLogger(topic));
    }

    /**
     * Shortcut that returns the LevelFactory
     */
    public static LevelFactory getLevelFactory() {
        return (LevelFactory) getLoggerFactory();
    }

    /**
     * Shortcut to get the Logger by its topic name.
     * @param is the topic of the returned logger
     * @return always a logger instance (never null value).
     */
    public static Logger getLogger(String topic) {
        return getLoggerFactory().getLogger(topic);
    }

    public static Properties getProperties() {
        return props;
    }
}
