/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * 00/01/08 Markus Fritz
 *          make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning
 *	    them from the pool.
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          For closing correctly all connections at server shutdown.
 * 01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism.
 *
 * 02/01/15 Dean Jennings - Map   instead of Hashtable  for cmList
 *
 * 03/01/13 Adriana Danes - Manage datasource name additionally to the
 *                          jndi name for the datasource using this ConnectionManager
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAResource;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.management.ReconfigDispatcher;
import org.objectweb.jonas.management.ReconfiguredProp;
import org.objectweb.transaction.api.ResourceManagerEvent;
import org.objectweb.transaction.api.ResourceManagerEventListener;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JOnAS generic implementation of DataSource to manage an underlying
 * JDBC-XA driver (XADataSource, XAConnection, XAResource, ...)
 * This class acts as a factory for Connection objects.
 * Its main goal is to manage a pool of XAConnection objects, to avoid 
 * closing the actual Connection on DBMS, and to deal with distributed
 * transactions.
 * This DataSource is registered in JNDI (=> implements Referenceable)
 *
 * @author Philippe Durieux
 * <p>
 * Contributor(s): 
 * <p>
 * 00/01/08 Markus Fritz make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning them from the pool.
 * <p>
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          For closing correctly all connections at server shutdown.
 * <p>01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism.
 * <p>02/01/15 Dean Jennings - Map   instead of Hashtable  for cmList
 * <p>Adriana Danes
 * <p>03/01/13 - Manage datasource name additionally to the jndi name for the datasource using this ConnectionManager
 * <p>03/04/08 - ConnectionManager extends ReconfigDispatcher in order to send reconfiguration Notifications to ReconfigManager
 * <p>03/05/23 - Introduce pool size configuration
 */

public class ConnectionManager extends ReconfigDispatcher 
    implements DataSource, 
    Referenceable, 
    Serializable,
    ConnectionEventListener,
    ConnectionManagerMBean {

    static private Logger logger = null;

    // The static table of all the ConnectionManager objects
    // Each ResourceManager has its own ConnectionManager (i.e. DataSource)
    private static Map cmList = new HashMap();

    /**
     * TransactionManager object
     * @serial
     */
    private TransactionManager tm = null;

    /**
     * underlaying XA DataSource
     * @serial
     */
    private XADataSourceImpl xads = null;

    /**
     * The pool of XAConnections managed by this Connection Manager
     * @serial
     */
    private Pool pool = null;

    // ----------------------------------------------------------
    // properties
    // ----------------------------------------------------------

    /**
     * @serial for JNDI
     */
    private String dSName = null;

    /**
     * @serial datasource name
     */
    private String datasourceName;

    /**
     * @serial url for database
     */
    private String url = null;
    /**
     * @serial JDBC driver Class
     */
    private String className = null;
    /**
     * @serial default user
     */
    private String userName = null;
    /**
     * @serial default passwd
     */
    private String password = null;
    /**
     * @serial 
     */
    private String connchecklevel = null;
    /**
     * @serial 
     */
    private String connmaxage = null;
    /**
     * @serial 
     */
    private String connteststmt = null;
    /**
     * @serial 
     */
    private String minconpool = null;
    /**
     * @serial 
     */
    private String maxconpool = null;

    /**
     * @serial 
     */
    private int isolationLevel = -1;
    private String isolationStr = null;

    /**
     * @serial 
     */
    private String currentMapperName = null;

    /**
     * Name of the client (client case)
     * TODO : must be removed. A client can lookup a datasource
     */
    private boolean isClient = false;

    /**
     * MBean variable
     */
    private String xaName = null;
    private String dsDescription = null;

    private static ResourceManagerEventListener rmListener = null;

    /**
     * Value used as sequence number by reconfiguration notifications
     */
    long sequenceNumber = 0;

    long getSequenceNumber() {
        return ++sequenceNumber;
    }
    
    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for Factory
     */
    public ConnectionManager() throws Exception {
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));

        // Set the TransactionManager
        TransactionService ts = null;
        try {
            ts = (TransactionService) ServiceManager.getInstance().getTransactionService();
        } catch (Exception e) {
            throw e;
        }

        tm = ts.getTransactionManager();

        // set threadData as being the listener for this RM
        setRMEListener((ResourceManagerEventListener) tm);       

        // Create the underlaying JDBC-XA driver
        // We do not use JNDI here. (should we do ?)
        // AD : do this before the Pool creation in order to allow the Pool constructor to access xads
        xads = new XADataSourceImpl();
        pool = new Pool(this, xads);
    }

    /**
     * Constructor for the client case
     * @param isClient true if this is the client case
     * TODO : must be removed. A client can lookup a datasource
     */
    public ConnectionManager(boolean isClient) throws Exception { 
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));
        this.isClient = isClient;
        xads = new XADataSourceImpl();
    }

    /**
     * This manager is in the client case or not ?
     * @return boolean true if this is the client case
     * TODO : must be removed. A client can lookup a datasource
     */
    public boolean isClientCase() {
        return isClient;
    }

    // -----------------------------------------------------------------
    // Properties
    // -----------------------------------------------------------------

    public void setRMEListener(ResourceManagerEventListener rmel) {
        rmListener = rmel;
    }
  
    /**
     * @return jndi name of the datasource
     */
    public String _getDSName() {
        return dSName;
    }
    /**
     * @param s jndi name for the datasource 
     */
    public void _setDSName(String s) {
        logger.log(BasicLevel.DEBUG, "entering ");
        dSName = s;
        xads.setDataSourceName(s); // for getRMID!
        // Add it to the list
        cmList.put(s, this);
    }

    public String getUrl() {
        return url;
    }
    public void setUrl(String s) {
        url = s;
        // Forward it to JDBC-XA driver
        xads.setUrl(s);
    }

    public String getClassName() {
        return className;
    }
    public void setClassName(String s) throws ClassNotFoundException {
        className = s;
        // Forward it to JDBC-XA driver
        xads.setClassName(s);
    }

    public String getUserName() {
        return userName;
    }
    public void setUserName(String s) {
        userName = s;
        // Forward it to JDBC-XA driver
        xads.setUserName(s);
    }

    public String getPassword() {
        return password;
    }
    public void setPassword(String s) {
        password = s;
        // Forward it to JDBC-XA driver
        xads.setPassword(s);
    }

    public void setTransactionIsolation(String level) {
        if (level.equals("serializable")) {
            isolationLevel = Connection.TRANSACTION_SERIALIZABLE;
        } else if (level.equals("none")) {
            isolationLevel = Connection.TRANSACTION_NONE;
        } else if (level.equals("read_committed")) {
            isolationLevel = Connection.TRANSACTION_READ_COMMITTED;
        } else if (level.equals("read_uncommitted")) {
            isolationLevel = Connection.TRANSACTION_READ_UNCOMMITTED;
        } else if (level.equals("repeatable_read")) {
            isolationLevel = Connection.TRANSACTION_REPEATABLE_READ;
        } else {
            isolationStr = "default";
            return;
        }
        // Forward it to JDBC-XA driver
        isolationStr = level;
        xads.setTransactionIsolation(isolationLevel);
    }

    public String getTransactionIsolation() {
        return isolationStr;
    }

    public void setMapperName(String mappername) {
        currentMapperName = mappername;
    }

    public String getMapperName() {
        return currentMapperName;
    }

    /**
     * Configure the Connection pool. Called by the Container at init.
     * Configuration can be set in datasource.properties files.
     * @param connchecklevel JDBC connection checking level
     * @param connmaxage JDBC connection maximum age
     * @param connteststmt SQL query for test statement
     * @param minconpool Min size for the connection pool
     * @param maxconpool Max size for the connection pool
     * @throw ServiceException if could not create the initial items in the pool
     */
    public void poolConfigure(String connchecklevel,
                              String connmaxage,
                              String connteststmt,
                              String minconpool,
                              String maxconpool) throws ServiceException {

        // Keep String values for getReference
        this.connchecklevel = connchecklevel;
        this.connmaxage = connmaxage;
        this.connteststmt = connteststmt;
        this.minconpool = minconpool;
        this.maxconpool = maxconpool;

        // Configure pool
        pool.setConnCheckLevel((new Integer(connchecklevel)).intValue());
        // set times in milliseconds
        pool.setConnMaxAge((new Integer(connmaxage)).longValue());
        pool.setTestStatement(connteststmt);
        pool.setMinConnPool((new Integer(minconpool)).intValue());
        pool.setMaxConnPool((new Integer(maxconpool)).intValue());
        logger.log(BasicLevel.DEBUG, "ConnectionManager configured with:");
        logger.log(BasicLevel.DEBUG, "   jdbcConnCheckLevel  = " + pool.jdbcConnCheckLevel);
        logger.log(BasicLevel.DEBUG, "   jdbcConnMaxAge      = " + pool.jdbcConnMaxAge);
        logger.log(BasicLevel.DEBUG, "   jdbcTestStmt        = " + pool.jdbcTestStmt);
        logger.log(BasicLevel.DEBUG, "   minConPool          = " + pool.getMinConnPool());
        logger.log(BasicLevel.DEBUG, "   maxConPool          = " + pool.getMaxConnPool());
    }

    // -----------------------------------------------------------------
    // DataSource implementation
    // -----------------------------------------------------------------

    /**
     * Attempt to establish a database connection.
     *
     * @return a Connection to the database
     *
     * @exception java.sql.SQLException - if a database-access error occurs.
     */
    public Connection getConnection() throws SQLException {
        return getConnection(userName, password);
    }

    /**
     * Attempt to establish a database connection.
     *
     * @param user - the database user on whose behalf the Connection is being made
     * @param password - the user's password
     *
     * @return a Connection to the database
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public Connection getConnection(String username, String password) throws SQLException {
        XAConnection xac = null;
        Transaction tx = null;
        try {
            tx = tm.getTransaction();
        } catch (NullPointerException n) {
            // current is null: we are not in a JOnAS Server.
            logger.log(BasicLevel.ERROR, "ConnectionManager: should not be used outside a JOnAS Server");
        } catch (SystemException e) {
            logger.log(BasicLevel.ERROR, "ConnectionManager: getTransaction failed:"+e);
        }
        logger.log(BasicLevel.DEBUG, "Tx = " + tx);

        // Check if available XAConnection in the pool for this user
        PoolItem pi = pool.openConnection(username, (Transaction) tx);
        if (pi == null) {
            // No available XAConnection in the pool -> try to create a new one and increase pool size if maxConPool boundary no reached
            pi = pool.newConnection(username, (Transaction) tx);
            if (pi != null)
                logger.log(BasicLevel.DEBUG, "Created new PoolItem");
            else
                logger.log(BasicLevel.DEBUG, "New PoolItem is null");
            // get the xac associated to the new PoolItem
            xac = pi.xaConn;
            if (xac != null)
                logger.log(BasicLevel.DEBUG, "Got created XA Connection");
            else
                logger.log(BasicLevel.DEBUG, "XA Connectio null!");
        } else {
            // We found a reusable XAConnection
            // We are still registered as ConnectionEventListener
            xac = pi.xaConn;
        }

        Connection ret = xac.getConnection();
        // Enlist XAResource if we are actually in a transaction
        if (tx != null) {
            logger.log(BasicLevel.DEBUG, "PI.open=" + pi.open);
            if (pi.open == 1) { // Only if first/only thread
                try {
                    tx.enlistResource(xac.getXAResource());
                } catch (RollbackException e) {
                    // Although tx has been marked to be rolled back, 
                    // XAResource has been correctly enlisted.
                    logger.log(BasicLevel.WARN, "XAResource enlisted, but tx is marked rollback");
                } catch (Exception e) {
                    logger.log(BasicLevel.ERROR, "Cannot enlist XAResource:"+e);
                    logger.log(BasicLevel.ERROR, "Connection will not be enlisted in a transaction");
                    // should return connection in the pool XXX
                    throw new SQLException("Cannot enlist XAResource");
                }
                // We are inside a transaction: no autocommit!
                ret.setAutoCommit(false);
            }
        } else {
            // We are not in a transaction yet: we just make a pre-registration.
            ret.setAutoCommit(true);    // in case we do not start a Tx
            rmListener.connectionOpened(pi);
        }
	
        // return a Connection object
        return ret;
    }

    /**
     * Get the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages 
     * printed by methods of other objects manufactured by this object, and so on.
     * Messages printed to a data source specificlog writer are not printed to the 
     * log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     *
     * @return the log writer for this data source, null if disabled
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public java.io.PrintWriter getLogWriter() throws SQLException {
        return xads.getLogWriter();
    }


    /**
     * Set the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages printed
     * by methods of other objects manufactured by this object, and so on. 
     * Messages printed to a data source specific log writer are not printed to 
     * the log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     * 
     * @param out - the new log writer; to disable, set to null
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public void setLogWriter(java.io.PrintWriter out) throws SQLException {
        xads.setLogWriter(out);
    }


    /**
     * Sets the maximum time in seconds that this data source will wait while 
     * attempting to connect to a database.
     * A value of zero specifies that the timeout is the default system timeout 
     * if there is one; 
     * otherwise it specifies that there is no timeout. 
     * When a data source object is created the login timeout is initially zero.
     *
     * @param seconds - the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public void setLoginTimeout(int seconds) throws SQLException {
        xads.setLoginTimeout(seconds);
    }

    /**
     * Gets the maximum time in seconds that this data source can wait while attempting 
     * to connect to a database.
     * A value of zero means that the timeout is the default system timeout if there is one; 
     * otherwise it means that there is no timeout. When a data source object is created 
     * the login timeout is initially zero.
     *
     * @return the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public int getLoginTimeout() throws SQLException {
        return xads.getLoginTimeout();
    }

    // -----------------------------------------------------------------
    // ConnectionEventListener implementation
    // -----------------------------------------------------------------

    /**
     * Invoked when the application calls close() on its representation 
     * of the connection.
     * @param event - an event object describing the source of the event
     */
    public void connectionClosed(ConnectionEvent event) {
        logger.log(BasicLevel.DEBUG, " ");

        XAConnection xac = (XAConnection) event.getSource();
        PoolItem pi = pool.closeConnection(xac, XAResource.TMSUCCESS);
	
        // remove it from the list of open connections for this thread
        // only if it was opened outside a tx.
        if (pi != null && pi.tx == null) {
            rmListener.connectionClosed(pi);
        }
    }

    /**
     * Invoked when a fatal connection error occurs, just before an
     * SQLException is thrown to the application.
     * @param event - an event object describing the source of the event
     */
    public void connectionErrorOccurred(ConnectionEvent event) {
        logger.log(BasicLevel.DEBUG, "entering ");

        XAConnection xac = (XAConnection) event.getSource();
        PoolItem pi = pool.closeConnection(xac, XAResource.TMFAIL);

        // remove it from the list of open connections for this thread
        // only if it was opened outside a tx.
        if (pi != null && pi.tx == null) {
            rmListener.connectionErrorOccured(pi);
        }
    }

    // -----------------------------------------------------------------
    // Referenceable implementation
    // -----------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while 
     * retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

        Reference ref = new Reference (this.getClass().getName(), 
                                       "org.objectweb.jonas.dbm.DataSourceFactory",
                                       null);
        // These values are used by ObjectFactory (see DataSourceFactory.java)
        ref.add(new StringRefAddr("datasource.name", getDSName()));
        ref.add(new StringRefAddr("datasource.url", getUrl()));
        ref.add(new StringRefAddr("datasource.classname", getClassName()));
        ref.add(new StringRefAddr("datasource.username", getUserName()));
        ref.add(new StringRefAddr("datasource.password", getPassword()));
        ref.add(new StringRefAddr("datasource.isolationlevel", getTransactionIsolation()));
        ref.add(new StringRefAddr("datasource.mapper", getMapperName()));
        ref.add(new StringRefAddr("connchecklevel", connchecklevel));
        ref.add(new StringRefAddr("connmaxage", connmaxage));
        ref.add(new StringRefAddr("connteststmt", connteststmt));
        return ref;
    }

    /**
     * get the ConnectionManager matching the DataSource name
     */
    public static ConnectionManager getConnectionManager(String dsname) {
       
        ConnectionManager cm = (ConnectionManager) cmList.get(dsname);
        return cm;
    }

    public void closeAllConnection(){
        logger.log(BasicLevel.DEBUG, "entering ");
        pool.closeAllConnections();
    }

    public static ResourceManagerEventListener getResourceManagerEventListener() {
        return rmListener;
    }

    /**
     * MBean method:
     * @return Integer number of open connection in this ds
     */
    public Integer getCurrentNumberOfJDBCConnectionOpen() {
        return new Integer(pool.getCurrentNumberOfXAConnection());
    }

    /**
     * MBean method:
     * @return Jndi name of the datasource
     */
    public String getDSName() {
        return dSName;
    }

    /**
     * @param s Jndi name of the datasource
     */
    public void setDSName(String s) {
        dSName = s;
    }
    
    /**
     * @return datasource name
     */
    public String getDatasourceName() {
        return datasourceName;
    }

    /**
     * @param s datasource name
     */
    public void setDatasourceName(String s) {
        datasourceName = s;
    }

    /**
     * MBean method:
     * @return the data source factory
     * none or JonasStandardXADataSource
     */
    public String getDataSourceFactory() {
        return "none";
    }

    /**
     * Management method:
     * @param the DataSourceXAName
     */
    public void setDataSourceXADataName(String xaName) {
        this.xaName = xaName;
    } 

    /**
     * MBean method:
     * @return the DataSourceXAName
     */
    public String getDataSourceXADataName() {
        return xaName;
    }
 
    /**
     * MBean method:
     * @return connection level
     */
    public Integer getJDBCConnCheckLevel() {
        return new Integer(connchecklevel);
    }

    /**
     * MBean method allowing to set the JDBC connection checking level value
     * @param connection level
     */
    public void setJDBCConnCheckLevel(Integer level) {
        pool.setConnCheckLevel(level.intValue());
        // The reconfigured property name is 'jdbc.connchecklevel'
        String propName = DataBaseServiceImpl.CONNCHECKLEVEL;
        // the new value is given by the 'level' argument
        ReconfiguredProp prop = new ReconfiguredProp(propName, level.toString());
        // Send a reconfiguration notification to the listner MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return connection max age
     */
    public Integer getJDBCConnMaxAge(){
        return new Integer(connmaxage);
    }
    /**
     * MBean method allowing to set the JDBC connection maximum age
     * @param connection max age
     */
    public void setJDBCConnMaxAge(Integer age) {
        connmaxage = age.toString();
        pool.setConnMaxAge(age.longValue());
        // The reconfigured property name is 'jdbc.connmaxage'
        String propName = DataBaseServiceImpl.CONNMAXAGE;
        // the new value is given by the 'age' argument
        ReconfiguredProp prop = new ReconfiguredProp(propName, age.toString());
        // Send a reconfiguration notification to the listner MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return min size of connection pool
     */
    public Integer getJDBCMinConnPool() {
        return new Integer(minconpool);
    }
    /**
     * MBean method allowing to set the minimum size of connection pool
     * @param 
     */
    public void setJDBCMinConnPool(Integer min) {
        logger.log(BasicLevel.DEBUG, "entering in setJDBCMinConnPool with " + min);
        minconpool = min.toString();
        // This opeartion has no effect on the current pool. It will be possibly taken
        // into account, if saveed, if the datasource is recreated (currently when the 
        // dbm service will restart, i.e. when JOnAS will restart 
        // ------
        // Update the configuration file
        // The reconfigured property name is 'jdbc.minconpool'
        String propName = DataBaseServiceImpl.MINCONPOOL;
        // the new value is given by the 'min' argument
        ReconfiguredProp prop = new ReconfiguredProp(propName, min.toString());
        // Send a reconfiguration notification to the listner MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return max size of connection pool
     */
    public Integer getJDBCMaxConnPool() {
        return new Integer(maxconpool);
    }
    /**
     * MBean method allowing to set the maximum size of connection pool
     * @param 
     */
    public void setJDBCMaxConnPool(Integer max) {
        logger.log(BasicLevel.DEBUG, "entering in setJDBCMaxConnPool with " + max);
        int ret = pool.setMaxConnPool(max.intValue());
        String s = (new Integer(ret)).toString();
        if (!s.equals(maxconpool)) {
            maxconpool = s;
        }
        // The reconfigured property name is 'jdbc.maxconpool'
        String propName = DataBaseServiceImpl.MAXCONPOOL;
        // the new value is given by the 'max' argument
        ReconfiguredProp prop = new ReconfiguredProp(propName, max.toString());
        // Send a reconfiguration notification to the listner MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return test statement 
     */
    public String getJDBCTestStatement(){
        return connteststmt;
    }

    /**
     * MBean method allowing to set the SQL query for test statement
     * @param test statement 
     */
    public void setJDBCTestStatement(String test) {
        connteststmt = test;
        pool.setTestStatement(test);
        // The reconfigured property name is 'jdbc.connteststmt'
        String propName = DataBaseServiceImpl.CONNTESTSTMT;
        // the new value is given by the 'test' argument
        ReconfiguredProp prop = new ReconfiguredProp(propName, test);
        // Send a reconfiguration notification to the listner MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }
    
    /**
     * Mbean Method:
     * return the type of this data source
     */
    public String getDataSourceType() {
        String result = null;
        if (getClassName().equals("oracle.jdbc.driver.OracleDriver")) result = "Oracle";
        if (getClassName().equals("interbase.interclient.Driver")) result = "InterBase";
        if (getClassName().equals("org.enhydra.instantdb.jdbc.idbDriver")) result = "Instant DB";
        if (getClassName().equals("postgresql.Driver")) result = "Postgres";     // For V. 6.x
        if (getClassName().equals("org.postgresql.Driver")) result = "Postgres"; // For V. 7.x
        return result;
    }

    /**
     * MBean Method:
     * @return the desrciption of this datasource
     */
    public String getDataSourceDescription() {
        return dsDescription;
    }

    /**
     * Management Method:
     * @param the desrciption of this datasource
     */
    public void setDataSourceDescription(String dsDesc) {
        dsDescription = dsDesc;
    }


    // MBean method not avaible in this type of connection
    /**
     * MBean method:
     * @return null
     */
    public Integer getJDBCSleepTimeOut(){
        return null;
    }
    /**
     * MBean method: do nothing 
     * @param timeout
     */
    public void setJDBCSleepTimeOut(Integer timeOut){
        // do nothing for this Connection type
    }
    /**
     * MBean method:
     * @return null
     */
    public Integer getJDBCDeadLockPool(){
        return null;
    }
    /**
     * MBean method: do nothing 
     * @param size
     */
    public void setJDBCDeadLockPool(Integer size){
        // do nothing for this Connection type
    }
    /**
     * MBean method:
     * @return null
     */
    public Integer getJDBCLoopWaitPool(){
        return null;
    }
    /**
     * MBean method: do nothing 
     * @param size
     */
    public void setJDBCLoopWaitPool(Integer size){
        // do nothing for this Connection type
    }

    /**
     * save updated configuration
     */
    public void saveConfig() { 
        sendSaveNotification(getSequenceNumber(), datasourceName);
    }
}
