/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.dbm;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
// JDBC imports
import javax.sql.XAConnection;
//
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAResource;
// JOnAS
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
// monolog
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Connection Pool
 * @author Philippe Durieux
 * <p>
 * Contributor(s): 
 * <ol>
 * <li>Markus Fritz</li>
 * <li>01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism to remove ThreadData
 *          dependency.</li>
 * <li>02/01/15 Dean Jennings - Map  instead of Hashtable for tx2item and xac2item</li>
 * <li>03/05/15 Adriana Danes - Introduce pool size configuration</li>
 * </ol>
 */

public class Pool {

    static private Logger logger = null;

    private TransactionManager tm = null;

    // List of free available items
    private LinkedList freeList = new LinkedList();
    // requests with no tx are identified by thread
    private int htcount = 0;
    // Requests with same tx get always the same conn
    private Map tx2item = new HashMap();
    // map XAConn => Item (to speedup CloseConn)
    private Map xac2item = new HashMap();

    // ---------------------------------------------------------------------
    // Pool Configuration. 
    // Values are initialized directly from ConnectionManager
    // ---------------------------------------------------------------------
    // Max age of a Connection in milliseconds.
    long jdbcConnMaxAge;
    // Level of checking on connections
    int jdbcConnCheckLevel;
    // test statement
    String jdbcTestStmt;
    // minimum size of the connection pool
    private int minConPool;
    // maximum size of the connection pool, -1  means no max boundary (default value)
    private int maxConPool;
    private boolean configured = false; // when created, the pool is not yet configured

    // AD - variables needed for supporting decerease of maxConPool
    // - reconfiguration opearation set maxConPool to a lower value that the current one
    private boolean toDecreaseMax = false;
    // - nb. of items to remove from the pool
    private int nbToDecrease;
    // - the updated value of maxConPool
    private int maxConPoolUpdate;

    // the pool's connection manager
    private ConnectionManager cmgr;
    private String datasourceName;
    // the XA datasource managed by the connection manager
    private XADataSourceImpl xads;

    /**
     * Pool constructor
     * @param xads XA DataSource implementation created by this pool's Connection manager. Allows
     * for getting XA Connections.
     * @param cmgr the pool's Connection manager has to be registered as ConnectionEventListener of the XA connections
     * created at the pool creation
     */
    public Pool(ConnectionManager cmgr, XADataSourceImpl xads) throws Exception {
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        logger.log(BasicLevel.DEBUG, "Create the Pool");
        this.cmgr = cmgr;
        this.xads = xads;
        this.datasourceName = cmgr.getDatasourceName();
        TransactionService ts = null;
        try {
            ts = (TransactionService) ServiceManager.getInstance().getTransactionService();
        } catch (Exception e) {
            throw e;
        }
        tm = ts.getTransactionManager();
        (new PoolKeeper(this)).start();
    }

    // Pool configuration methods
    // --------------------------
    /**
     *
     */
    protected int setConnCheckLevel(int jdbcConnCheckLevel) {
       this.jdbcConnCheckLevel = jdbcConnCheckLevel;
       return this.jdbcConnCheckLevel;
    }
    /**
     * @param jdbcConnMaxAge max age of connection in minutes
     */
    protected long setConnMaxAge(long jdbcConnMaxAge) {
       // set times in milliseconds
       this.jdbcConnMaxAge = jdbcConnMaxAge * 60L * 1000L;
       // return the parameter (in minutes)
       return jdbcConnMaxAge;
    }
    /**
     *
     */
    protected String setTestStatement(String jdbcTestStmt) {
       this.jdbcTestStmt = jdbcTestStmt;
       return this.jdbcTestStmt;
    }
    /**
     *
     */
    protected int setMinConnPool(int minConPool) throws ServiceException {
        this.minConPool = minConPool;
        if (minConPool > 0) {          
            for (int i = 0 ; i < minConPool ; i++) {
                XAConnection xac = null;
                try {
                    xac = xads.getXAConnection();
                } catch (SQLException e) {
                    throw new ServiceException("Could not create " + minConPool + " items in the pool (" + i + " were created): " + e.toString());
                }
                // tx = null, jdbcConnMaxAge already configured (see ConnectionManager.poolConfigure() method)
                PoolItem item = new PoolItem(this, xac, null, jdbcConnMaxAge);
                freeList.addLast(item);
                xac2item.put(xac, item);
                xac.addConnectionEventListener(cmgr);
            }
        }
        logger.log(BasicLevel.DEBUG, "Pool configured with minConPool = " + this.minConPool);
        return this.minConPool;
    }
    /**
     * 
     */
    protected int setMaxConnPool(int maxConPool) {
        if (!configured) {
            // this is an initialization opeartion
            this.maxConPool = maxConPool;
            logger.log(BasicLevel.DEBUG, "Pool configured with maxConPool = " + this.maxConPool);
            configured = true;
        } else {
            // this is a reconfiguration opeartion
            // 0 - no change
            // 1 - increase
            // 2 - decrease
            int opCode = 0;
            if (this.maxConPool == -1) {
                // decrease from infinite to maxConPool
                opCode = 2;
            } else {
                if (maxConPool == -1) {
                    // increase from this.maxConPool to infinite
                    opCode = 1;
                } else {
                    // maxConPool et this.maxConPool are both positif
                    if (maxConPool > this.maxConPool) {
                        // increase the pool size
                        opCode = 1;
                    } else {
                        // decrease the pool size
                        opCode = 2;
                    }
                }
            }

            switch (opCode) {
            case 0 : 
                break;
            case 1 : 
                if (toDecreaseMax) {
                    // Cancel previous decrease reconfiguration
                    toDecreaseMax = false;
                }
                this.maxConPool = maxConPool;
                logger.log(BasicLevel.DEBUG, "Done increaseing maxConPool to " + this.maxConPool);
                break;
            case 2 :
                // decrease the pool size
                logger.log(BasicLevel.DEBUG, "Reconfiguring pool: decreaseing the maxConPool to " + maxConPool);
                int currentNumberOfConnections = xac2item.size();
                if (currentNumberOfConnections < maxConPool) {
                    // can decrease imediatelly
                    this.maxConPool = maxConPool;
                    logger.log(BasicLevel.DEBUG, "Done decreaseing maxConPool to " + this.maxConPool);
                } else {
                    // cannot decrease for the moment, wait till the number of connections decreases enought
                    toDecreaseMax = true;
                    nbToDecrease = currentNumberOfConnections - maxConPool;
                    maxConPoolUpdate = maxConPool;
                    logger.log(BasicLevel.WARN, "Datasource + " + datasourceName + " configuration: Cannot currently decrease upper limit of connection pool (number of connections to destroy is: " + nbToDecrease + ")");
                }
                break;
            }
        }
        return this.maxConPool;
    }

    protected int getMaxConnPool() {
        return this.maxConPool;
    }

    protected int getMinConnPool() {
        return this.minConPool;
    }  
 
    /**
     * lookup connection in the pool for this user/tx
     * @param user user name
     * @param tx Transaction the connection is involved
     * @return the free PoolItem or null if no item available
     * @throw SQLException throwed if cannot open a connection because the pool's max size is reached
     */
    public synchronized PoolItem openConnection(String user, Transaction tx) throws SQLException {
        logger.log(BasicLevel.DEBUG, "");
        PoolItem item = null;
        // Try to find a connection for this tx.
        if (tx != null) {
            Object i;
            while ((i = tx2item.get(tx)) != null) {
                logger.log(BasicLevel.DEBUG, "tx2item containes item for this tx");
                if (i instanceof PoolItem) {
                    item = (PoolItem)i;
                    logger.log(BasicLevel.DEBUG, "this is a real item");
                    break;
                }
                // Wait for another thread to call newConnection().
                logger.log(BasicLevel.DEBUG, "this is a placeholder");
                try {
                    wait();
                    if (tx2item.get(tx) == null) {
                        logger.log(BasicLevel.DEBUG, "I'v got notified, but failed to get an item");
                        // the newConnection() failed (the pool boundary reached)
                        throw new SQLException("The connection pool of " + cmgr.getDatasourceName() + " datasource has reached the maximum size");
                    } // else, the while condition gets evaluated and i gets the item to be returned
                    logger.log(BasicLevel.DEBUG, "I'v got notified and I'll go back in the while test to get the item");
                } catch (InterruptedException ign) {
                }
            }
        }
        if (item == null) {
            // No connection found for this tx, or no tx.
            // try to find an item in the free list
            logger.log(BasicLevel.DEBUG, "no connection found for this tx, or no tx");
            while (true) {
                if (freeList.isEmpty()) {
                    logger.log(BasicLevel.DEBUG, "empty free list");
                    if (tx != null) {
                        // Put a placeholder before returning null in case of multiple threads
                        // Assumes this thread will call newConnection() after.
                        logger.log(BasicLevel.DEBUG, "create a placeholder");
                        tx2item.put(tx, new Object());
                    }
                    return null;
                }
                logger.log(BasicLevel.DEBUG, "remove an PI from the freeList");
                item = (PoolItem) freeList.removeFirst();

                // Check the connection before reusing it
                if (jdbcConnCheckLevel > 0) {
                    try {
                        logger.log(BasicLevel.DEBUG, "Check the JDBC connection");
                        Connection conn = item.xaConn.getConnection();
                        if (conn.isClosed()) {
                            destroyItem(item);
                            continue;
                        }
                        if (jdbcConnCheckLevel > 1) {
                            logger.log(BasicLevel.DEBUG, "retrying connection: "+jdbcTestStmt);
                            java.sql.Statement stmt = conn.createStatement();
                            stmt.execute(jdbcTestStmt);
                            stmt.close();
                        }
                    } catch (Exception e) {
                        logger.log(BasicLevel.ERROR, "Datasource + " + datasourceName + " error: removing invalid item:" + e);
                        destroyItem(item); 
                        continue;
                    }
                }
                break;
            }
            // We found an item. Associate it with the tx.
            logger.log(BasicLevel.DEBUG, "Found an item. Associate it with the tx");
            item.tx = tx;
            if (tx == null) {
                htcount++;
            } else {
                //logger.log(BasicLevel.DEBUG, "reuse a Connection for Tx = "+tx);
                tx2item.put(tx, item);
                // register synchronization
                try {
                    tx.registerSynchronization(item);
                } catch (javax.transaction.RollbackException e) {
                    /// optimization is probably possible at this point
                    logger.log(BasicLevel.WARN, "Datasource + " + datasourceName + " error: Pool item registered, but tx is rollback only");
                } catch (javax.transaction.SystemException e) {
                    logger.log(BasicLevel.ERROR, "Datasource + " + datasourceName + " error in pool: system exception from transaction manager " + e.getMessage());
                }
            }
        }

        // TODO : waiting for exclusive connection ?

        item.open++;
        return item;
    }

    /**
     * Add a new XAConnection in the pool
     * @param user user name
     * @param tx Transaction the connection is involved
     * @throw SQLException throwed if cannot open a connection because the pool's max size is reached
     */
    public synchronized PoolItem newConnection(String user, Transaction tx) throws SQLException {
        // if the pool has no upper boundary or if a max boundary exists but it is not yet reached, than create a new connection
        if ((maxConPool <0) || (maxConPool > 0 && xac2item.size() < maxConPool)) {
            logger.log(BasicLevel.DEBUG, "Create a PI and a Connection for Tx = " + tx);
            // create a new XA Connection
            XAConnection xac = xads.getXAConnection();
            logger.log(BasicLevel.DEBUG, "Create a new PoolItem");
            PoolItem item = new PoolItem(this, xac, user, jdbcConnMaxAge);
            // Register the connection manager as a ConnectionEventListener
            xac.addConnectionEventListener(cmgr);
            logger.log(BasicLevel.DEBUG, "add the item to xac2item");
            xac2item.put(xac, item);	
            item.tx = tx;
            item.open++;
            if (tx == null) {
                htcount++;
            } else {
                // replace the placeholder by the new item
                logger.log(BasicLevel.DEBUG, "Remove the placeholder for TX");
                tx2item.remove(tx);
                logger.log(BasicLevel.DEBUG, "Insert the new item for TX");
                tx2item.put(tx, item);
                // register synchronization
                try {
                    tx.registerSynchronization(item);
                } catch (javax.transaction.RollbackException e) {
                    // optimization is probably possible at this point
                    logger.log(BasicLevel.WARN, "Poolitem registered, but tx is rollback only");
                } catch (javax.transaction.SystemException e) {
                    logger.log(BasicLevel.ERROR, "Pool: system exception from transaction manager " + e.getMessage());
                }
                notifyAll();
            }
            logger.log(BasicLevel.DEBUG, "Return item");
            return item;
        } else {
            if (tx != null) {
                // remove the  placeholder
                tx2item.remove(tx);
                // notify the possible ones waiting to receive an item for this tx
                notifyAll();
            }
            throw new SQLException("The connection pool of " + cmgr.getDatasourceName() + " datasource has reached the maximum size");
        }
    }
   
    /**
     * The transaction has committed (or rolled back). We can return its
     * connections to the pool of available connections.
     * @param tx the non null transaction
     */
    public synchronized void freeConnections(Transaction tx) {
        logger.log(BasicLevel.DEBUG, "Free connection for Tx = " + tx);
        PoolItem item = (PoolItem) tx2item.remove(tx);
        if (item == null) {
            logger.log(BasicLevel.ERROR, "Pool: no connection found to free for Tx = " + tx);
            return;
        }
        item.tx = null;
        if (item.open > 0) {
            // Forcing a physical close to avoid reusing it in another transaction may
            // lead to deadlock in database (close on a connection currently in use)
            //item.open = 0;
            //destroyItem(item);
            // XXX don't know what is the best to do here ?
            logger.log(BasicLevel.ERROR, "WARNING: Connection not closed by caller");
            return;
        }
        freeItem(item);
    }

    /**
     * Mark a specific Connection in the pool as closed.
     * If it is no longer associated to a Tx, we can free it.
     *
     * @return the PoolItem the connection is in (=> tx)
     * or null if error.
     */
    public synchronized PoolItem closeConnection(XAConnection xac, int flag) {
        PoolItem item = (PoolItem) xac2item.get(xac);
        if (item == null) {
            logger.log(BasicLevel.ERROR, "Pool: no item for this xac!");
            return null;
        }
        // The connection will be available only if not associated
        // to a transaction. Else, it will be reusable only for the
        // same transaction.
        item.open--;
        if (item.open < 0) {
            item.open = 0;
            logger.log(BasicLevel.WARN, "connection was already closed");
            return null;
        }
        if (item.tx == null && item.open > 0) {
            logger.log(BasicLevel.ERROR, "connection-open counter overflow");
            item.open = 0;
        }
        if (item.tx != null) {
            logger.log(BasicLevel.DEBUG, "keep the Connection for tx = "+item.tx);
        } else {
            logger.log(BasicLevel.DEBUG, "free the Connection");
            htcount--;
            // Add it to the free list if pool is not currently shrinking
            // else, remove the item from the pool
            freeItem(item);
        }
        // TODO notify item here ???

        // delist Resource if in transaction
        Transaction tx = null;
        try {
            tx = tm.getTransaction();
        } catch (NullPointerException n) {
            // current is null: we are not in JOnAS Server.
            logger.log(BasicLevel.ERROR, "Pool: should not be used outside a JOnAS Server");
        } catch (SystemException e) {
            logger.log(BasicLevel.ERROR, "Pool: getTransaction failed:" + e);
        }
        if (tx != null && item.open < 1) {
            try {
                tx.delistResource(xac.getXAResource(), flag);
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Pool: Exception while delisting resource:"+e);
            }
        }

        return item;
    }

    // This method replaces freeList.addLast(item) instruction
    // which was enoght before when pool resizing was no supported
    private void freeItem(PoolItem item) {        
        if (!toDecreaseMax) {
            // Add it to the free list as before
            freeList.addLast(item);
            logger.log(BasicLevel.DEBUG, "Item added to freeList");
        } else {
            if (nbToDecrease > 0) {
                // this item is to be forgotten, not added in the freeList
                nbToDecrease--;
                logger.log(BasicLevel.DEBUG, "Item removed");
                logger.log(BasicLevel.DEBUG, "Continue decreasing the pool size with " + nbToDecrease + "(in progress)");
            } else {
                // OK, now the pool is sized, do delayed maxConPool update 
                toDecreaseMax = false;
                this.maxConPool = maxConPoolUpdate;
                freeList.addLast(item);
                logger.log(BasicLevel.DEBUG, "Item added to freeList");
                logger.log(BasicLevel.DEBUG, "Done decreaseing maxConPool to " + this.maxConPool);
            }
        }
    }

    /**
     * Check all connections not in a transaction if their maxage
     * is reached and if so kill them.
     */
    public synchronized void check4MaxAge() {
        PoolItem item;
        Iterator i = freeList.iterator();
        while (i.hasNext()) {
            item = (PoolItem) i.next();
            if (item.tx != null || item.open > 0) {
                logger.log(BasicLevel.ERROR, "Inconsistent Pool item in FreeList!");
            }
            if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                logger.log(BasicLevel.DEBUG, "remove a timed out Connection");
                i.remove();
                destroyItem(item);
            }
        }
    }

    /**
     * Close all connections in the pool, when server is shut down.
     */
    public synchronized void closeAllConnections() {
        Iterator enum = xac2item.keySet().iterator();
        try {
            while (enum.hasNext()) {
                XAConnection xac = (XAConnection) enum.next();
                xac.close();
            }
        } catch (java.sql.SQLException e) {
        }
    }

    /**
     * Destroy an item because connection closed or error occured
     */
    private void destroyItem(PoolItem item) { 
        logger.log(BasicLevel.DEBUG, "remove an item in xac2item");       
        xac2item.remove(item.xaConn);	
        try {
            item.xaConn.close();
        } catch (java.sql.SQLException ign) {
        }
        item.xaConn = null;
    }

    /**
     * Management method:
     * @return int number of xa connection
     */
    public int getCurrentNumberOfXAConnection() {
        return xac2item.size();
    }
}
