/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Jeff Mesnil.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security;

import java.util.Hashtable;
import java.security.Guard;
import java.security.Principal;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Check if the principal role is matching the role defined for the method
 * @author Jeff Mesnil (initial developper)
 * @author Florent Benoit 04/03 check if there are roles in SecurityContext
 */

public class MethodGuard implements Guard, java.io.Serializable {

    /**
     * Logger for debug
     */
    private static Logger logger = null;
    
    /**
     * list of method signatures identifying the methods to guard.
     */
    private Hashtable methods = new Hashtable();

    /**
     * Associate method guard to a container
     */
    private String containerName = null;

    /**
     * Constructor
     * @param containerName name of the container
     */
    public MethodGuard (String containerName) {
        this.containerName = containerName;
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }
    }

    /**
     * The roles which can access to the method
     * @param methodSignature signature of the method
     * @param roles roles required to execute themethod
     */
    public void add(String methodSignature, String[] roles) {
        logger.log(BasicLevel.DEBUG, "entering for " + methodSignature);
        methods.put(methodSignature, roles);
    }
    

    /**
     * Check if the role(s) of the principal match the roles of the method
     * @param methodSignature signature of the method
     * @throws SecurityException if the access is denied
     */
    public void checkGuard(Object methodSignature) throws SecurityException {

        // Not authorized by default
        boolean isAuthorized = false;
        
        // Roles of the method
        String[] methodRoles = (String[]) methods.get(methodSignature);

        // No roles, everyone can access the method
        if (methodRoles.length == 0) {
            logger.log(BasicLevel.DEBUG, "access to method not protected");
            return;
        }

        // Verify if principal has at least one of the roles
        SecurityCurrent current = SecurityCurrent.getCurrent();
        SecurityContext ctx = current.getSecurityContext();
        if (ctx == null) {
            String err = "no security context exists";
            logger.log(BasicLevel.ERROR, err);
            throw new SecurityException(err);
        }


        Principal principal = ctx.getCallerPrincipal();
        String[] principalRoles = ctx.getCallerPrincipalRoles();
        RoleMechanism roleMechanism =  RoleMechanism.getInstance();
        if (principalRoles == null) {
            isAuthorized = roleMechanism.isCallerInRoles(principal, methodRoles);
        } else {
            isAuthorized = roleMechanism.isCallerInRoles(principal, methodRoles, principalRoles);
        }
        if (!isAuthorized) {
            String err = "not authorized to access Method";
            logger.log(BasicLevel.DEBUG, err);
            throw new SecurityException(err);
        }
        logger.log(BasicLevel.DEBUG, "authorized to access Method");
        return;
    }

    /**
     * Method toString
     * @return String a string representation of the object
     */
    public String toString() {
        return "MethodGuard of container : " + containerName;
    }
}
