/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Jeff Mesnil.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security;

import java.security.Principal;
import java.security.Guard;
import java.util.Hashtable;
import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Check access for the role-link/role-ref
 * @author Jeff Mesnil (initial developper)
 * @author Florent Benoit 04/03 check if there are roles in SecurityContext
 *
 */

public class RoleGuard implements Guard {

    /**
     * Logger for debug
     */
    private static Logger logger = null;
    
    /**
     * Associate role guard to a container
     */
    private String containerName = null;

    /**
     * Mapping between a role-ref and a role
     */
    private Hashtable roleRefs = new Hashtable();

    /**
     * Constructor
     * @param containerName name of the container
     */
    public RoleGuard(String containerName) {
        this.containerName = containerName;
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }
    }

    /**
     * Associate a role-ref to a role
     * @param roleRef the role to associate
     * @param role the associated role
     */
    public void add(String roleRef, String role) {
        logger.log(BasicLevel.DEBUG, "mapping between roleref ='" + roleRef + "' and role '" + role + "'.");
        roleRefs.put(roleRef, role);
    }

    /**
     * Check if the role(s) of the principal match the role-ref
     * @param roleString the role
     * @throws SecurityException if the access is denied
     */
    public void checkGuard(Object roleString) throws SecurityException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "role " + (String) roleString);
        }

        // Not authorized by default
        boolean isCallerInRole = false;

        String roleRef = (String) roleString;
        String role = (String) roleRefs.get(roleRef);
        if (role == null) {
            logger.log(BasicLevel.DEBUG, roleRef + " has not been defined");
            throw new SecurityException("the logical role associated with " + roleRef + " has not been defined");
        }
        RoleMechanism roleMechanism = RoleMechanism.getInstance();
        SecurityCurrent current = SecurityCurrent.getCurrent();
        SecurityContext ctx = current.getSecurityContext();
        if (ctx == null) {
            String err = "no security context exists";
            logger.log(BasicLevel.DEBUG, err);
            throw new SecurityException(err);
        }
        Principal principal = ctx.getCallerPrincipal();
        String[] principalRoles = ctx.getCallerPrincipalRoles();

        if (principalRoles == null) {
            isCallerInRole = roleMechanism.isCallerInRole(principal, role);
        } else {
            isCallerInRole = roleMechanism.isCallerInRole(principal, role, principalRoles);
        }

        if (isCallerInRole) {
            logger.log(BasicLevel.DEBUG, "caller is in role " + roleString);
            return;
        } else {
            String err = "Caller is not in role " + roleString;
            logger.log(BasicLevel.DEBUG, err);
            throw new SecurityException(err);
        }
    }

    /**
     * Method toString
     * @return String a string representation of the object
     */
    public String toString() {
        return "RoleGuard of container : " + containerName;
    }
}
