/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Jeff Mesnil.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security;

import java.security.Principal;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Enumeration;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.security.realm.factory.JResource;
import org.objectweb.jonas.security.realm.factory.JResourceException;
import org.objectweb.jonas.security.realm.principals.User;

/**
 * Managed roles defined in a resource
 * and check the right access between method roles and principal roles
 * @author Jeff Mesnil
 * @author Florent Benoit : use a JOnAS Resource (Memory, Datasource, LDAP)
 */

public class RoleMechanism {

    /**
     * The resource we will use to authenticate users
     * and identify associated roles.
     */
    private JResource jResource = null;

    /**
     * Logger for debug
     */
    private static Logger logger = null;

    /**
     * Unique instance
     */
    private static RoleMechanism unique = null;
    
    /**
     * Associate a principal name to a list of roles (String[])
     */
    private Hashtable principals;

    /**
     * Default constructor
     */
    private RoleMechanism() {
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }
        principals = new Hashtable();
    }

    /**
     * Return the unique instance
     * @return RoleMechanism the unique instance
     */
    public static RoleMechanism getInstance() {
        if (unique == null) {
            unique = new RoleMechanism();
        }
        return unique;
    } 

    /**
     * Use the given realm for the mapping between the principal and the roles
     * @param resourceName the realm to use
     * @throws Exception if the realm doesn't exist
     */
    public void setRealmResource(String resourceName) throws Exception {

        if (resourceName == null) {
            logger.log(BasicLevel.ERROR, "No realm set in jonas.properties file.");
            return;
        }

        SecurityService securityService = null;
        // Get the Security Service
        try {
            securityService = (SecurityService) ServiceManager.getSecurityService();
        } catch (ServiceException e) {
            // Can't retrieve Security service
            throw new Exception("Can't retrieve Security service");
        }
        // Get the resource from the security service
        jResource = securityService.getJResource(resourceName);
        if (jResource == null) {
            throw new Exception("Can't retrieve resource " + resourceName + "from the security service");
        }

    }

    /**
     * Check if the given role match the principal roles
     * @param principal principal
     * @param role the given role
     * @param pRoles the roles of the principal
     * @return boolean true if given role match the principal roles
     */
    public boolean isCallerInRole(Principal principal, String role, String[] pRoles) {
        logger.log(BasicLevel.DEBUG, "entering for principal " + principal + ", role : " + role);
        boolean isCallerInRole = false;
        String[] principalRoles = null;

        if (pRoles != null) {
            principalRoles = pRoles;
        } else {
            principalRoles = getArrayRolesFromPrincipalName(principal.getName());
        }

        if (principalRoles == null) {
            return false;
        }
        for (int i = 0; i < principalRoles.length; i++) {
            isCallerInRole = isCallerInRole || (role.equals(principalRoles[i]));
            if (isCallerInRole) {
                break;
            }
        }
        return isCallerInRole;
    }        

    /**
     * Check if the given role match the principal roles
     * @param principal principal
     * @param role the given role
     * @return boolean true if given role match the principal roles
     */
    public boolean isCallerInRole(Principal principal, String role) {
        return isCallerInRole(principal, role, null);
    }

    /**
     * Check if the given roles match the principal roles
     * @param principal principal
     * @param roles the given roles
     * @param pRoles the roles of the principal
     * @return boolean if given roles match the principal roles
     */
    public boolean isCallerInRoles(Principal principal, String[] roles, String[] pRoles) {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "principal: " + principal);
        }

        boolean isCallerInRole = false; 
        logger.log(BasicLevel.DEBUG, this);

        String[] principalRoles = null;
        if (pRoles != null) {
            principalRoles = pRoles;
        } else {
            principalRoles = getArrayRolesFromPrincipalName(principal.getName());
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "principalRoles : " + principalRoles); 
        }
        if (principalRoles != null) {
            for (int i = 0; i < roles.length; i++) { 
                for (int j = 0; j < principalRoles.length; j++) {
                    isCallerInRole = isCallerInRole || (roles[i].equals(principalRoles[j]));
                }
                if (isCallerInRole) {
                    break;
                }
            }
        }
        if (isCallerInRole) {
            logger.log(BasicLevel.DEBUG, "Caller is authorized to access the method");
        } else {
            logger.log(BasicLevel.DEBUG, "Caller is not authorized to access the method");
        }
        return isCallerInRole;


    }

    /**
     * Check if the given roles match the principal roles
     * @param principal principal
     * @param roles the given roles
     * @return boolean if given roles match the principal roles
     */
    public boolean isCallerInRoles(Principal principal, String[] roles) {
        return isCallerInRoles(principal, roles, null);
    }

    /**
     * Method toString
     * @return String a string representation of the object
     */
    public String toString() {
        StringBuffer buff = new StringBuffer();
        buff.append("RoleMechanismImpl : ");
        Enumeration keys = principals.keys();
        while (keys.hasMoreElements()) {
            String  principalName = (String) keys.nextElement();
            buff.append("\n\t" + principalName);
            String[] roles = (String[]) principals.get(principalName);
            buff.append(", roles : ");
            for (int i = 0; i < roles.length; i++) {
                buff.append(roles[i] + ", ");
            }
        } 
        return new String(buff);
    }

    private String[] getArrayRolesFromPrincipalName(String username) {

        // No authentication can be made with a null username
        if (username == null) {
            logger.log(BasicLevel.DEBUG, "No username so no authentication");
            return null;
        }

        // Does a user with this username exist?
        User user = null;
        try {
            user = jResource.findUser(username);
        } catch (Exception jre) {
            // could not retrieve user
            logger.log(BasicLevel.INFO, "Can not find the user : " + jre.getMessage());
            return null;
        }
        
        // User was not found
        if (user == null) {
            logger.log(BasicLevel.DEBUG, "User " + username + " not found.");
            return null;
        }

        ArrayList combinedRoles = null;
        try {
            combinedRoles = jResource.getArrayListCombinedRoles(user);
        } catch (JResourceException jre) {
            logger.log(BasicLevel.ERROR, jre.getMessage());
            return null;
        }
        
        // Convert ArrayList into array
        String[] r = new String[combinedRoles.size()];
        r = (String[]) combinedRoles.toArray(r);

        return r;
    }


}
