/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security.auth.spi;

// import java
import java.util.ArrayList;
import java.util.Map;

// import javax
import javax.naming.InitialContext;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;
import javax.security.auth.Subject;

// import jonas
import org.objectweb.jonas.security.realm.factory.JResource;
import org.objectweb.jonas.security.realm.factory.JResourceException;
import org.objectweb.jonas.security.realm.principals.User;

import org.objectweb.jonas.security.auth.JGroup;
import org.objectweb.jonas.security.auth.JPrincipal;
import org.objectweb.jonas.security.auth.JRole;

/**
 * Define a login module for the authentication by using one of the JOnAS resource
 * Datasource, LDAP or file
 */
public class JResourceLoginModule {

    /**
     * Subject used
     */
    private Subject subject = null;

    /**
     * The callbackhandler user for identification
     */
    private CallbackHandler callbackHandler = null;

    /**
     * Shared state with all login modules
     */
    private Map sharedState = null;

    /**
     * Options for this login module
     */
    private Map options = null;

    /**
     * Name of the principal
     */
    private String principalName = null;

    /**
     * Roles of the principal
     */
    private ArrayList principalRoles = null;

    /**
     * Initialize this LoginModule.
     * This method is called by the LoginContext after this LoginModule has been instantiated. The purpose of this method is to initialize this LoginModule with the relevant information. If this LoginModule does not understand any of the data stored in sharedState or options parameters, they can be ignored.
     * @param subject the Subject to be authenticated.
     * @param callbackHandler a CallbackHandler for communicating with the end user (prompting for usernames and passwords, for example).
     * @param sharedState state shared with other configured LoginModules.
     * @param options options specified in the login Configuration for this particular LoginModule.
     */
    public void initialize(Subject subject, CallbackHandler callbackHandler, Map sharedState, Map options) {
        this.subject = subject;
        this.callbackHandler = callbackHandler;
        this.sharedState = sharedState;
        this.options = options;
    }


    /**
     * Method to authenticate a Subject (phase 1).
     * The implementation of this method authenticates a Subject. For example, it may prompt for Subject information such as a username and password and then attempt to verify the password. This method saves the result of the authentication attempt as private state within the LoginModule.
     * @return true if the authentication succeeded, or false if this LoginModule should be ignored. 
     * @throws LoginException if the authentication fails
     */
    public boolean login() throws LoginException {

        // No handler
        if (callbackHandler == null) {
            throw new LoginException("No handler has been defined.");
        }

        // Resource to be used (jndi name)
        String resourceName = (String) options.get("resourceName");

        // No resource is specified -> fail
        if (resourceName == null) {
            throw new LoginException("You have to give an argument to this login module. The 'resourceName' parameter is required.");
        }

        // Get the resource and perform authentication
        try {
            InitialContext ictx = new InitialContext();
            JResource jResource = null;
            try {
                jResource = (JResource) ictx.lookup(resourceName);
            } catch (Exception e) {
                throw new LoginException("Can not retrieve the resource '" + resourceName + "'. Check that this resource is bound in the registry.");
            }
           
            // Handle callbacks
            NameCallback nameCallback = new NameCallback("User :");
            // False to hide password when it is entered
            PasswordCallback passwordCallback = new PasswordCallback("Password :", false);
            Callback[] callbacks = new Callback[] {nameCallback, passwordCallback};
            callbackHandler.handle(callbacks);

            // Get values
            principalName = nameCallback.getName();
            if (principalName == null) {
                throw new LoginException("A null username is not a valid username");
            }
            char[] arrayPass = passwordCallback.getPassword();
            if (arrayPass == null) {
                throw new LoginException("A null password is not a valid password");
            }
            String password = new String(arrayPass);

            // Authentication - step 1 (user)
            User user = null;
            try {
                user = jResource.findUser(principalName);
            } catch (Exception jre) {
                // could not retrieve user
                throw new LoginException("Can not find the user : " + jre.getMessage());
            }
            // User was not found
            if (user == null) {
                throw new LoginException("User '" + principalName + "' not found.");
            }

            // Authentication - step 2 (password)
            boolean validated = jResource.isValidUser(user, password);
            if (!validated) {
                throw new LoginException("The password for the user '" + principalName + "' is not valid");
            }

            // Authentication - step 3 (roles)
            try {
                principalRoles = jResource.getArrayListCombinedRoles(user);
            } catch (JResourceException jre) {
                 throw new LoginException(jre.getMessage());
            }

        } catch (Exception e) {
            throw new LoginException("Error during the login phase : " + e.getMessage());
        }


        return true;
    }


    /**
     * Method to commit the authentication process (phase 2).
     * This method is called if the LoginContext's overall authentication succeeded (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules succeeded).
     * If this LoginModule's own authentication attempt succeeded (checked by retrieving the private state saved by the login method), then this method associates relevant Principals and Credentials with the Subject located in the LoginModule. If this LoginModule's own authentication attempted failed, then this method removes/destroys any state that was originally saved.
     * @return true if this method succeeded, or false if this LoginModule should be ignored. 
     * @throws LoginException if the commit fails
     */
    public boolean commit() throws LoginException {

        // Add principal to the current subject
        subject.getPrincipals().add(new JPrincipal(principalName));
        JGroup group = new JGroup("Roles");

        //Convert list into array
        String[] roles = new String[principalRoles.size()];
        roles = (String[]) principalRoles.toArray(roles);

        for (int i = 0; i < principalRoles.size(); i++) {
            group.addMember(new JRole(roles[i]));
        }

        // Add group
        subject.getPrincipals().add(group);

        return true;
    }
    

    /**
     * Method to abort the authentication process (phase 2).
     * This method is called if the LoginContext's overall authentication failed. (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules did not succeed).
     * If this LoginModule's own authentication attempt succeeded (checked by retrieving the private state saved by the login method), then this method cleans up any state that was originally saved.
     * @return true if this method succeeded, or false if this LoginModule should be ignored.
     * @throws LoginException if the abort fails
     */
    public boolean abort() throws LoginException {
        
        // Reset temp values
        principalName = null;
        principalRoles = null;

        return true;
    }

    /**
     * Method which logs out a Subject.
     * An implementation of this method might remove/destroy a Subject's Principals and Credentials.
     * @return true if this method succeeded, or false if this LoginModule should be ignored. 
     * @throws LoginException if the logout fails
     */
    public boolean logout() throws LoginException {

        // Remove principal name
        subject.getPrincipals().remove(new JPrincipal(principalName));

        return true;
    }

}
