/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Craig R. McClanahan (Jakarta Tomcat) for RealmBase class.
 *
 * The Apache Software License, Version 1.1
 * Copyright (c) 1999 The Apache Software Foundation.  All rights reserved.
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security.realm;

//import java
import java.util.ArrayList;
import java.security.Principal;


//import catalina
import org.apache.catalina.LifecycleException;
import org.apache.catalina.realm.RealmBase;
import org.apache.catalina.realm.GenericPrincipal;
import org.apache.catalina.realm.Constants;
import org.apache.catalina.util.StringManager;

//import objectweb
import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.jonas.common.Log;

//import monolog
import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.security.SecurityService;
import org.objectweb.jonas.security.realm.factory.JResource;
import org.objectweb.jonas.security.realm.factory.JResourceException;
import org.objectweb.jonas.security.realm.principals.User;

/**
 * <p>Implementation of a Realm. (by a wrapper)
 * Use any JOnAS realm by specifying the resource name
 *
 * @author Craig R. McClanahan
 * @author Florent Benoit : Tomcat 4.1.x /JOnAS 3.1
 */
public class JRealmCatalina41 extends RealmBase {

    /**
     * The resource we will use to authenticate users
     * and identify associated roles.
     */
    private JResource jResource = null;


    /**
     * Descriptive information about this Realm implementation.
     */
    private final String info = "org.objectweb.jonas.security.realm.JRealmCatalina41/1.0";


    /**
     * Descriptive information about this Realm implementation.
     */
    private static final String name = "JRealmCatalina41";


    /**
     * The name of the resource
     */
    private String resourceName = null;


    /**
     * The string manager for this package.
     */
    private static StringManager sm = StringManager.getManager(Constants.Package);
    
    /**
     * Reference to the JOnAS security service
     */
    private SecurityService securityService = null;


    /**
     * The logger used in JOnAS
     */
    private static Logger logger = null;


    /**
     * Return descriptive information about this Realm implementation and
     * the corresponding version number, in the format
     * <code>&lt;description&gt;/&lt;version&gt;</code>.
     * @return the info.
     */
    public String getInfo() {
        return info;
    }


    /**
     * Return the resource name we will be using.
     * @return the resource name.
     */
    public String getResourceName() {
        return resourceName;
    }


    /**
     * Set the resource name we will be using.
     *
     * @param resourceName The new resource name
     */
    public void setResourceName(String resourceName) {
        this.resourceName = resourceName;

    }


    /**
     * Return the Principal associated with the specified username and
     * credentials, if there is one; otherwise return <code>null</code>.
     *
     * @param username Username of the Principal to look up
     * @param credentials Password or other credentials to use in
     *                    authenticating this username
     * @return the principal associated
     */
    public Principal authenticate(String username, String credentials) {

        // No authentication can be made with a null username
        if (username == null) {
            logger.log(BasicLevel.DEBUG, "No username so no authentication");
            return null;
        }

        // Does a user with this username exist?
        User user = null;
        try {
            user = jResource.findUser(username);
        } catch (Exception jre) {
            // could not retrieve user
            logger.log(BasicLevel.INFO, "Can not find the user : " + jre.getMessage());
            return null;
        }
        
        // User was not found
        if (user == null) {
            logger.log(BasicLevel.DEBUG, "User " + username + " not found.");
            return null;
        }

        boolean validated = jResource.isValidUser(user, credentials);
        if (!validated) {
            logger.log(BasicLevel.INFO, "The password for the user " + username + " is not valid");
            return null;
        }

        ArrayList combinedRoles = null;
        try {
            combinedRoles = jResource.getArrayListCombinedRoles(user);
        } catch (JResourceException jre) {
            logger.log(BasicLevel.ERROR, jre.getMessage());
            return null;
        }

        GenericPrincipal principal = new GenericPrincipal(this, user.getName(), user.getPassword(), combinedRoles);
        SecurityContext ctx = new SecurityContext(principal.getName(), combinedRoles);
        SecurityCurrent current = SecurityCurrent.getCurrent();
        current.setSecurityContext(ctx);

        return principal;
    }


    /**
     * Return a short name for this Realm implementation.
     * @return the name
     */
    protected String getName() {
        return name;
    }


    /**
     * Return the password associated with the given principal's user name.
     * @param username the given principal's user name.
     * @return the password associated.
     */
    protected String getPassword(String username) {
        return null;
    }


    /**
     * Return the Principal associated with the given user name.
     * @param username the given principal's user name.
     * @return the Principal associated
     */
    protected Principal getPrincipal(String username) {
        return null;
    }


    // ------------------------------------------------------ Lifecycle Methods


    /**
     * Prepare for active use of the public methods of this Component.
     *
     * @exception LifecycleException if this component detects a fatal error
     *  that prevents it from being started
     */
    public synchronized void start() throws LifecycleException {

        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }

        // Get the Security Service
        try {
            securityService = (SecurityService) ServiceManager.getSecurityService();
        } catch (ServiceException e) {
            // Can't retrieve Security service
            throw new LifecycleException("can't retrieve Security service");
        }
        
        // Get the resource from the security service
        jResource = securityService.getJResource(resourceName);
        if (jResource == null) {
            throw new LifecycleException("Can't retrieve resource " + resourceName + "from the security service");
        }

        // Perform normal superclass initialization
        super.start();

    }


    /**
     * Gracefully shut down active use of the public methods of this Component.
     *
     * @exception LifecycleException if this component detects a fatal error
     *  that needs to be reported
     */
    public synchronized void stop() throws LifecycleException {
        // Perform normal superclass finalization
        super.stop();

        // Release reference to our resource
        jResource = null;
    }

   /**
     * Log a message on the Logger associated with our Container (if any)
     *
     * @param message Message to be logged
     */
    protected void log(String message) {
        logger.log(BasicLevel.DEBUG, message);
    }



}
