/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Greg Wilkins for the HashUserRealm
 * Copyright (c) 1996 Mort Bay Consulting Pty. Ltd. All rights reserved.
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security.realm;

//import java
import java.util.ArrayList;


//import objectweb
import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.jonas.common.Log;

//import monolog
import org.objectweb.util.monolog.api.BasicLevel;

import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.security.SecurityService;
import org.objectweb.jonas.security.realm.factory.JResource;
import org.objectweb.jonas.security.realm.principals.User;

import org.mortbay.http.UserRealm;
import org.mortbay.http.UserPrincipal;
import org.mortbay.http.HttpRequest;

/**
 * <p>Implementation of a Realm.
 * Use any JOnAS realm by specifying the resource name
 * @author Greg Wilkins for the HashUserRealm
 * @author Florent Benoit : Jetty 4.2.x / JOnAS 3.1
 */
public class JRealmJetty42 implements UserRealm {

    /**
     * Name of this realm
     */
    private String name;

    /**
     * The resource we will use to authenticate users
     * and identify associated roles.
     */
    private JResource jResource = null;

    /**
     * The name of the resource
     */
    private String resourceName = null;


    /**
     * The logger used in JOnAS
     */
    private static Logger logger = null;

    /**
     * Reference to the JOnAS security service
     */
    private SecurityService securityService = null;


    /**
     * Constructor
     * @param resourceName name of the resource to use
     * @throws Exception if the resource can't be retrieved
     */
    public JRealmJetty42(String resourceName) throws Exception {
        this.resourceName = resourceName;

        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_SECURITY_PREFIX);
        }

        // Get the Security Service
        try {
            securityService = (SecurityService) ServiceManager.getSecurityService();
        } catch (ServiceException e) {
            // Can't retrieve Security service
            throw new Exception("can't retrieve Security service");
        }
        
        // Get the resource from the security service
        jResource = securityService.getJResource(resourceName);
        if (jResource == null) {
            throw new Exception("Can't retrieve resource " + resourceName + "from the security service");
        }
    }

    /**
     * Constructor
     * @param name name of the realm
     * @param resourceName name of the resource to use
     * @throws Exception if the resource can't be retrieved
     */
    public JRealmJetty42(String name, String resourceName) throws Exception {
        this(resourceName);
        this.name = name;
    }


    /** 
     * @return The realm name. 
     */
    public String getName() {
        return name;
    }

    /**
     * Authenticate a user with a specific username and credentials
     * @param username name of the user
     * @param credentials credential of the user
     * @param request httprequest
     * @return a Jetty principal
     */
    public UserPrincipal authenticate(String username, Object credentials, HttpRequest request) {
        // No authentication can be made with a null username
        if (username == null) {
            return null;
        }

        // Does a user with this username exist?
        User user = null;
        try {
            user = jResource.findUser(username);
        } catch (Exception jre) {
            // could not retrieve user
            logger.log(BasicLevel.INFO, jre.getMessage());
            return null;
        }
        
        // User was not found
        if (user == null) {
            logger.log(BasicLevel.DEBUG, "User " + username + " not found.");
            return null;
        }

        if (!(credentials instanceof String)) {
            logger.log(BasicLevel.ERROR, "Allow only string type as credentials");
            return null;
        }


        boolean validated = jResource.isValidUser(user, (String) credentials);

        if (!validated) {
            logger.log(BasicLevel.INFO, "The password for the user " + username + " is not valid");
            return null;
        }

        ArrayList combinedRoles = null;
        try {
            combinedRoles = jResource.getArrayListCombinedRoles(user);
        } catch (Exception jre) {
            logger.log(BasicLevel.ERROR, jre.getMessage());
            return null;
        }

        JettyPrincipal principal = new JettyPrincipal(user.getName(), user.getPassword(), combinedRoles);
        SecurityContext ctx = new SecurityContext(principal.getName(), combinedRoles);
        SecurityCurrent current = SecurityCurrent.getCurrent();
        current.setSecurityContext(ctx);
        return principal;
    }

    /**
     * Disassociate a user
     * Not implemented
     * @param user the given user
     */
    public void disassociate(UserPrincipal user) {
        // Not implemented
    }
    
    /**
     * Push a role to a user
     * Not implemented
     * @param user the given user
     * @param role the role to push
     * @return the new principal
     */
    public UserPrincipal pushRole(UserPrincipal user, String role) {
        return user;
    }

    /**
     * Pop a role to a user
     * Not implemented
     * @param user the given user
     * @return the new principal
     */
    public UserPrincipal popRole(UserPrincipal user) {
        return user;
    }


    /**
     * Define a principal which can be used by Jetty
     */
    private class JettyPrincipal implements UserPrincipal {
       
        /**
         * Name of the principal
         */
        private String name;
        
        /**
         * Credential of the principal
         */
        private String credential;

        /**
         * Roles for this principal
         */
        private ArrayList roles;

        /**
         * Constructor
         * @param name name of this principal
         * @param credential credential of the principal
         * @param roles roles for this principal
         */
        public JettyPrincipal(String name, String credential, ArrayList roles) {
            this.name = name;
            this.credential = credential;
            this.roles = roles;
        }

        /**
         * Get the name of this principal
         * @return the name of this principal
         */
        public String getName() {
            return name;
        }

        /**
         * This user is authenticated ?
         * @return true if it is authenticated
         */
        public boolean isAuthenticated() {
            return true;
        }
        
        /**
         * Check if the given role is in the user's roles
         * @param role the given role
         * @return true if the role is in the user's roles
         */
        public boolean isUserInRole(String role) {
            if (roles == null) {
                return (false);
            }
            return roles.contains(role);
        }
       
    }

}



