/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.security.realm.factory;

import java.security.NoSuchAlgorithmException;
import java.util.Hashtable;
import java.util.ArrayList;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.StringRefAddr;

import java.sql.Connection;
import javax.sql.DataSource;
import java.sql.DriverManager; 
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import javax.naming.Context;
import javax.naming.InitialContext;

import org.objectweb.jonas.dbm.ConnectionManager;

import org.objectweb.jonas.security.realm.principals.User;
import org.objectweb.jonas.security.realm.lib.HashHelper;

//import monolog
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class extends the JResource class for the Datasource implementation.
 * @author Florent Benoit
 */
public class JResourceDS extends JResource implements JResourceDSMBean {

    /**
     * Type of the factory
     */
    private static final String FACTORY_TYPE = "org.objectweb.jonas.security.realm.factory.JResourceDS";

    /**
     * Name of the factory
     */
    private static final String FACTORY_NAME = "org.objectweb.jonas.security.realm.factory.JResourceDSFactory";


    /**
     * Name of the datasource resource to use.
     */
    private String dsName = null;

    /**
     * Name of table which have the username/password
     */
    private String userTable = null;

    /**
     * Column of the username of the user table
     */
    private String userTableUsernameCol = null;

    /**
     * Column of the password of the user table
     */
    private String userTablePasswordCol = null;

    /**
     * Name of table which have the username/role
     */
    private String roleTable = null;

    /**
     * Column of the username of the role table
     */
    private String roleTableUsernameCol = null;

    /**
     * Column of the role of the role table
     */
    private String roleTableRolenameCol = null;


    /**
     * Define the user statement
     */
    private PreparedStatement userStatement = null;

    /**
     * Define the role statement
     */
    private PreparedStatement roleStatement = null;

    /**
     * Datasource to use
     */
    private DataSource dataSource = null;


    /**
     * Constructor. Use the super constructor
     * @throws Exception if super constructor fail
     */
    public JResourceDS() throws Exception {
        super();
        
    }

    /**
     * Set the name of the resource to use
     * @param dsName name of the resource
     */
    public void setDsName(String dsName) {
        this.dsName = dsName;
    }

    /**
     * Set the name of the table which have the username/password
     * @param userTable name of the table which have the username/password
     */
    public void setUserTable(String userTable) {
        this.userTable = userTable;
    }

    /**
     * Set the name of the column of the username of the user table
     * @param userTableUsernameCol name of the column of the username of the user table
     */
    public void setUserTableUsernameCol(String userTableUsernameCol) {
        this.userTableUsernameCol = userTableUsernameCol;
    }

    /**
     * Set the name of column of the password of the user table
     * @param userTablePasswordCol name of column of the password of the user table
     */
    public void setUserTablePasswordCol(String userTablePasswordCol) {
        this.userTablePasswordCol = userTablePasswordCol;
    }

    /**
     * Set the name of table which have the username/role
     * @param roleTable name of table which have the username/role
     */
    public void setRoleTable(String roleTable) {
        this.roleTable = roleTable;
    }

    /**
     * Set the name of the column of the username of the role table
     * @param roleTableUsernameCol name of the column of the username of the role table
     */
    public void setRoleTableUsernameCol(String roleTableUsernameCol) {
        this.roleTableUsernameCol = roleTableUsernameCol;
    }

    /**
     * Set the name of the column of the role of the role table
     * @param roleTableRolenameCol name of the column of the role of the role table
     */
    public void setRoleTableRolenameCol(String roleTableRolenameCol) {
        this.roleTableRolenameCol = roleTableRolenameCol;
    }

    /**
     * Get the name of the resource to use
     * @return name of the resource
     */
    public String getDsName() {
        return dsName;
    }

    /**
     * Get the name of the table which have the username/password
     * @return name of the table which have the username/password
     */
    public String getUserTable() {
        return userTable;
    }

    /**
     * Get the name of the column of the username of the user table
     * @return name of the column of the username of the user table
     */
    public String getUserTableUsernameCol() {
        return userTableUsernameCol;
    }

    /**
     * Get the name of column of the password of the user table
     * @return name of column of the password of the user table
     */
    public String getUserTablePasswordCol() {
       return userTablePasswordCol;
    }

    /**
     * Get the name of table which have the username/role
     * @return name of table which have the username/role
     */
    public String getRoleTable() {
        return roleTable;
    }

    /**
     * Get the name of the column of the username of the role table
     * @return name of the column of the username of the role table
     */
    public String getRoleTableUsernameCol() {
        return roleTableUsernameCol;
    }

    /**
     * Get the name of the column of the role of the role table
     * @return name of the column of the role of the role table
     */
    public String getRoleTableRolenameCol() {
       return roleTableRolenameCol;
    }


    /**
     * Check if a user is found and return it
     * @param username the wanted user name
     * @return the user found or null
     * @throws JResourceException if there is a SQLException
     */
    public User findUser(String username) throws JResourceException {

        if (username == null) {
            return null;
        }

        // Look in the cache
        User user = (User) users.get(username);
        
        if (user != null) {
            return user;
        }

        Connection connection = getConnection();

        user = new User();
        user.setName(username);

        // Get the password of the user
        ResultSet rs = null;
        String password = null;
        try {
            rs = userStatement(connection, username).executeQuery();
            rs.last();
            int records = rs.getRow();
            if (records > 1) {
                logger.log(BasicLevel.ERROR, "There are more than one user with the name" + username);
            }
            password = rs.getString(userTablePasswordCol).trim();
            rs.close();
        } catch (SQLException sqle) {
            closeConnection(connection);
            throw new JResourceException(sqle.getMessage());
        }

        if (password == null) {
            closeConnection(connection);
            return null;
        }
        user.setPassword(password);
       
        // Get the roles of the user
        try {
            rs = roleStatement(connection, username).executeQuery();
            while (rs.next()) {
                String role = rs.getString(roleTableRolenameCol).trim();
                user.addRole(role);
            }
            rs.close();
        } catch (SQLException sqle) {
            closeConnection(connection);
            throw new JResourceException(sqle.getMessage());
        }
        closeConnection(connection);
        return user;
    }


    /**
     * Check if the given credential is the right credential for the given user
     * @param user user to check its credentials
     * @param credentials the given credentials
     * @return true if the credential is valid for this user
     */
    public boolean isValidUser(User user, String credentials) {

        // Look in the cache. Validated if in cache
        if (users.get(user.getName()) != null) {
            return true;
        }

        boolean validated = false;

        //Get algorithm and hashpassword
        String pass = user.getHashPassword().getPassword();
        String algo = user.getHashPassword().getAlgorithm();

        // Crypt password ?
        if (algo != null) {
            try {
                validated = HashHelper.hashPassword(credentials, algo).equalsIgnoreCase(pass);
            } catch (NoSuchAlgorithmException nsae) {
                logger.log(BasicLevel.ERROR, "Can't make a password with the algorithm " + algo + ". " + nsae.getMessage());
            }
        } else {
            // clear
            validated = credentials.equals(pass);
        }
        return validated;
    }

    /**
     * Get all the roles (from the roles and from the groups) of the given user
     * @param user the given user
     * @return the array list of all the roles for a given user
     * @throws JResourceException if it fails
     */
    public ArrayList getArrayListCombinedRoles(User user) throws JResourceException {

        // Look in the cache. get roles from the cache
        User userFound = (User) users.get(user.getName());
        if (userFound != null) {
            return userFound.getCombinedRoles();
        }
       
        ArrayList allCombinedRoles = new ArrayList();

        // Return empty array if user null
        if (user == null) {
            return allCombinedRoles;
        }
        // Add all user roles
        String[] userRoles = user.getArrayRoles();
        for (int r = 0; r < userRoles.length; r++) {
            String roleName = userRoles[r];
            if (!allCombinedRoles.contains(roleName)) {
                allCombinedRoles.add(roleName);
            }
        }
        user.setCombinedRoles(allCombinedRoles);
        // add in cache
        users.put(user.getName(), user);
        return allCombinedRoles;
    }

    /**
     * String representation of the MemoryRealm
     * @return the xml representation of the MemoryRealm
     */
    public String toXML() {
        StringBuffer xml = new StringBuffer("    <dsrealm name=\"");
        xml.append(getName());
        xml.append("\"\n             dsName=\"");
        if (dsName != null) {
            xml.append(dsName);
        }
        xml.append("\"\n             userTable=\"");
        if (userTable != null) {
            xml.append(userTable);
        }
        xml.append("\" userTableUsernameCol=\"");
        if (userTableUsernameCol != null) {
            xml.append(userTableUsernameCol);
        }
        xml.append("\" userTablePasswordCol=\"");
        if (userTablePasswordCol != null) {
            xml.append(userTablePasswordCol);
        }
        xml.append("\"\n             roleTable=\"");
        if (roleTable != null) {
            xml.append(roleTable);
        }
        xml.append("\" roleTableUsernameCol=\"");
        if (roleTableUsernameCol != null) {
            xml.append(roleTableUsernameCol);
        }
        xml.append("\" roleTableRolenameCol=\"");
        if (roleTableRolenameCol != null) {
            xml.append(roleTableRolenameCol);
        }

        xml.append("\" />");
        return xml.toString();
    }


    /**
     * The string representation of this realm is the XML
     * @return XML representation
     */
    public String toString() {
        return this.toXML();
    }

    /**
     * Retrieves the Reference of the object. 
     * The Reference contains the factory used to create this object 
     * and the optional parameters used to configure the factory.
     * @return the non-null Reference of the object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public Reference getReference() throws NamingException {

        // Build the reference to the factory FACTORY_TYPE
        Reference reference = new Reference(FACTORY_TYPE, FACTORY_NAME, null);

        // Add ref addr
        reference.add(new StringRefAddr("name", getName()));
        reference.add(new StringRefAddr("dsName", dsName));
        reference.add(new StringRefAddr("userTable", userTable));
        reference.add(new StringRefAddr("userTableUsernameCol", userTableUsernameCol));
        reference.add(new StringRefAddr("userTablePasswordCol", userTablePasswordCol));
        reference.add(new StringRefAddr("roleTable", roleTable));
        reference.add(new StringRefAddr("roleTableUsernameCol", roleTableUsernameCol));
        reference.add(new StringRefAddr("roleTableRolenameCol", roleTableRolenameCol));

        return reference;
    }



    /**
     * Try to close the given connection
     * @param c the connection to close
     */
    private void closeConnection(Connection c) {
        if (c == null) {
            return;
        }
        try {
            c.close();
        } catch (Exception e) {
            logger.log(BasicLevel.DEBUG, "Can not close the connection");
        }

    }

    /**
     * Get a connection from the dataSource
     * @return  the connection from the dataSource
     * @throws JResourceException if an SQLException is thrown by dataSource.getConnection()
     */
    private Connection getConnection() throws JResourceException {
        // If no dataSource, get an instance
        if (dataSource == null) {
            // Finds DataSource from JNDI
            Context initialContext = null;
            try {
                initialContext = new InitialContext();
                dataSource = (DataSource) initialContext.lookup(dsName);
            } catch (Exception e) {
                String err = "Could not found resource " + dsName + " in the registry" + e.getMessage();
                logger.log(BasicLevel.ERROR, err);
                throw new JResourceException(err);
            }
        }


        // Check if the request come from a client or inside the server
        // Required until a datasource can be called from a client
        ConnectionManager cm = null;
        if (dataSource instanceof ConnectionManager) {
            cm = (ConnectionManager) dataSource;
        } else {
            throw new JResourceException("Only the JOnAS datasources can be used at this moment");
        }

        Connection c = null;
        
        // Client or server case ?
        if (cm.isClientCase()) {
            // Build a JDBC connection with getting parameters from the object ConnectionManager
            try {
                // Load driver
                Class.forName(cm.getClassName());
                // create the connection
                c = DriverManager.getConnection(cm.getUrl(), cm.getUserName(), cm.getPassword()); 
            } catch (Exception de) {
                logger.log(BasicLevel.ERROR, de.getMessage());
                throw new JResourceException("Can not build a connection using the jdbc parameters :" + de.getMessage());
            }
        } else {
            //Use the datasource
            // Retrieve connection from the datasource
            try {
                c = dataSource.getConnection();
            } catch (SQLException sqle) {
                logger.log(BasicLevel.ERROR, sqle.getMessage());
                throw new JResourceException(sqle.getMessage());
            }

        }

        return c;
    }


    /**
     * Return the user query. It select the password for a specific user
     * @return the user query
     */
    private String userQuery() {
        /* 
           SELECT userTablePasswordCol
           FROM userTable
           WHERE userTableUsernameCol = ?
        */ 
        StringBuffer stringBuffer = new StringBuffer("SELECT ");
        stringBuffer.append(userTablePasswordCol);
        stringBuffer.append(" FROM ");
        stringBuffer.append(userTable);
        stringBuffer.append(" WHERE ");
        stringBuffer.append(userTableUsernameCol);
        stringBuffer.append(" = ?");
        return (stringBuffer.toString());
    }

    /**
     * Return the roles query. It select the roles for a specific user
     * @return the roles query
     */
    private String rolesOfUserQuery() {
        /* 
           SELECT r.roleTableRolenameCol
           FROM userTable u, roleTable r
           WHERE u.userTableUsernameCol = r.roleTableUsernameCol
           AND u.userTableUsernameCol = ?
        */ 
        
        StringBuffer stringBuffer = new StringBuffer("SELECT r.");
        stringBuffer.append(roleTableRolenameCol);
        stringBuffer.append(" FROM ");
        stringBuffer.append(userTable);
        stringBuffer.append(" u, ");
        stringBuffer.append(roleTable);
        stringBuffer.append(" r WHERE u.");
        stringBuffer.append(userTableUsernameCol);
        stringBuffer.append(" = r.");
        stringBuffer.append(roleTableUsernameCol);
        stringBuffer.append(" AND u.");
        stringBuffer.append(userTableUsernameCol);
        stringBuffer.append(" = ?");
        return stringBuffer.toString();
    }

    /**
     * Return a statement for the given username by using the userQuery query
     * @param connection connection to use
     * @param username the given user
     * @return a statement for the given user
     * @throws SQLException if the SQL statement fails
     */
    private PreparedStatement userStatement(Connection connection, String username) throws SQLException {
        
        if (userStatement == null) {
            userStatement = connection.prepareStatement(userQuery());
        }
        
        userStatement.setString(1, username);
        return userStatement;
    }

    /**
     * Return a statement for the given username by using the rolesOfUserQuery query
     * @param connection connection to use
     * @param username the given user
     * @return a roles of user statement for the given user
     * @throws SQLException if the SQL statement fails
     */
    private PreparedStatement roleStatement(Connection connection, String username) throws SQLException {
        
        if (roleStatement == null) {
            roleStatement = connection.prepareStatement(rolesOfUserQuery());
        }
        
        roleStatement.setString(1, username);

        return roleStatement;
    }
    

    /**
     * Return all the users in the cache
     * @return all the users in the cache
     */
    public Hashtable getUsers() {
        return users;
    }

    /**
     * Clear the cache
     */
    public void clearCache() {
        users = new Hashtable();
    }

}
