/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.server;

import java.util.StringTokenizer;
import java.rmi.RMISecurityManager;
import javax.management.MBeanServer;

import org.objectweb.jonas.adm.Adm;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.common.LogManagement;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.management.ReconfigManager;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas_ejb.lib.Version;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;
import org.objectweb.carol.util.configuration.CarolConfiguration;

/**
 * This class implements an JOnAS Server.
 * @author Philippe Coq, Philippe Durieux
 * 00/06/15: Jeff Mesnil
 * 01/06/15: Regis Le Brettevillois - Libelis / JOnAS team - evidian
 *           Service as org.objectweb.jonas.service.Service
 * 02/04/25: Sebastien Chassande-Barrioz sebastien.chassande@inrialpes.fr
 *           Monolog for traces
 *
 *           Adriana Danes
 * 03/01/20:  -Highlight configuration properties
 * 03/03/..:  -Force jmx service startup
 -Create and register the JonAS configuration MBean. This MBean manages JOnAS persistent re-configuration (see ReconfigManager  class)
*/
public class Server {

    static private Logger logger = null;
    static private String SERVERNAME = "JOnAS"; // J2EE Application Server Name
    private String serverName = null; // Current server name (the server instance name)

    // The Server singleton value
    private static Server unique = null;

    // The associated ServiceManager for JOnAS
    ServiceManager sm = null;

    // Associated Adm object
    Adm adm = null;

    // JOnAS configuration properties
    JProp props = null;

    // Number of active threads in the initial thread's group
    int threadCount = 0;
    
    // Estimated number of active groups in the initial thread's group
    int groupCount = 0;

    // String protocols
    String protocols = null;

    // Server configuration properties
    static final String NAME = "jonas.name";
    static final String DEF_NAME = "jonas";
    static final String ORB_PORT = "jonas.orb.port";
    static final String DEF_ORB_PORT = "0";
    static final String SECURITY_PROPAGATION = "jonas.security.propagation";
    static final String TRANSACTION_PROPAGATION = "jonas.transaction.propagation";
    static final String LOG_CONFIGFILE = "jonas.log.configfile";
    static final String DEF_LOG_CONFIGFILE = "trace";
    static final String CONFIGFILE_USECLASSPATH = "jonas.log.configfile.useclasspath";
    static final String DEF_CONFIGFILE_USECLASSPATH = "true";

    // Server configuration properties
    int orbPort;
    String logConfigFile;
 
    /**
     * Private constructor to force only 1 instance.
     */
    private Server() throws Exception {
        // read the JOnAS properties
        props = JProp.getInstance();

        // set configuration info into the Log System (monolog)
        logConfigFile = props.getValue(LOG_CONFIGFILE, DEF_LOG_CONFIGFILE);
        Log.configure(logConfigFile);

        // get a logger for server traces
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

        // create the service manager
        sm = ServiceManager.getInstance();
    }

    /**
     * Get the unique instance.
     * Create it at first call.
     */
    public static Server getInstance() throws Exception {
        if (unique == null) {
            unique = new Server();
        }
        return unique;
    }

    /**
     * Start the server
     */
    public void start() throws Exception {

        // Server name
        serverName = props.getValue(NAME, DEF_NAME);
	
        // Initialize Context Interceptors and Carol
        String protocols = initCarol(props);

        // RMI Security Manager
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        // First of all, start registry service 
        sm.startRegistry();

        // then, start JMX service (start an MBean server and create RMI connector for it registered in JNDI)

        sm.startJmx();

        // get the JMX server
        JmxService jmxService = null;
        MBeanServer jmxServer = null;
        try {
            jmxService = (JmxService) ServiceManager.getJmxService();
            jmxServer = jmxService.getJmxServer();
        } catch (ServiceException se) {
            // this could mean that jmx service not available
            // this is not possible, as we started it even if it does not apper in the services list configuration parameter
            logger.log(BasicLevel.ERROR, "Could not access the jmx service");	
            throw se;
        }

        try {
            // create and register a MBean allowing server reconfiguration management
            ReconfigManager reconfigManagerMBean = new ReconfigManager(props.getConfigFileEnv(), jmxServer);
            jmxServer.registerMBean(reconfigManagerMBean, JonasObjectName.serverConfig());

            // create and register a MBean allowing the JOnAS server administration (currently only monitoring aspects)
            // TODO change JmxServer class name !!! 
            jmxServer.registerMBean(new JmxServer(this, serverName, Version.NUMBER, protocols), JonasObjectName.ejbServer());

            // create and register a MBean for the logging service administration
            LogManagement logMBean = LogManagement.getInstance();
            jmxServer.registerMBean(logMBean, JonasObjectName.logService(logConfigFile));
        } catch (Exception e) {
            // Cannot administrate this server throu JMX
            jmxService.stop();	    
        }

        // Creates the Adm instance for remote administration
        try {
            adm = new Adm(props);
        } catch (Exception e) {
            e.printStackTrace();
            throw new Exception("Cannot create Adm : "+e);
        }

        // start all other services
        sm.startServices();

        // TM Server or EJB Server ?
        boolean isEJB = true;
        try {
            Service ejbservice = ServiceManager.getEjbService();
        } catch (ServiceException se) {
            isEJB = false;
        }
        adm.serverReady(isEJB);
        System.out.println("The JOnAS Server '"+serverName+"' version "+Version.NUMBER+" is ready");
        logger.log(BasicLevel.INFO, "Jonas server "+serverName+" started on "+protocols);

        // Periodically count the number of active threads in this thread's group
        Thread thisThread = Thread.currentThread();
        ThreadGroup thGroup = thisThread.getThreadGroup();
        while(true) {	    
            threadCount = Thread.activeCount();
            groupCount = thGroup.activeGroupCount();
            thisThread.sleep(10000);
        }
    }

    /**
     * get the number of active threads in the JOnAS server's initial thread group
     */
    protected int getActiveThreadsCount() {
        return threadCount;
    }
    /** 
     * get the estimated number of active groups in the JOnAS server's initial thread group
     */
    public int getActiveGroupCount() {
        return groupCount;
    }

   /**
     * initialize interception and carol
     */
    public String  initCarol(JProp props) throws Exception {

        // Intialize Carol 
        CarolConfiguration.init(Log.getLoggerFactory());

        // Initialize Interceptors
        String sec = props.getValue(SECURITY_PROPAGATION, "false").trim();
        String trans = props.getValue(TRANSACTION_PROPAGATION, "false").trim();
        if (sec.equals("true")) {
            // jrmp interceptors
            CarolConfiguration.addInterceptors("jrmp", "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // cmi interceptors
            CarolConfiguration.addInterceptors("cmi", "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // iiop interceptors flag    
            CarolConfiguration.addInterceptors("iiop", "org.objectweb.jonas.security.interceptors.iiop.SecurityInitializer");
	    
        }
	
        if (trans.equals("true")) {
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("jrmp", "org.objectweb.jonas_tm.jta.rmi.JTAInterceptorInitializer");
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("cmi", "org.objectweb.jonas_tm.jta.rmi.JTAInterceptorInitializer");
            // iiop interceptors flag    
            CarolConfiguration.addInterceptors("iiop", "org.objectweb.jonas_tm.ots.OTSORBInitializer");
        }
	
	// intialize port number
	System.setProperty("rmi.server.port",props.getValue(ORB_PORT, DEF_ORB_PORT));

        String s = CarolConfiguration.getProtocols();
        StringTokenizer st = new StringTokenizer(s, ",", false);
        String pn = " ";

        int idx = 0;
        while (st.hasMoreElements()) {
            if (idx != 0) {
                pn = pn + ", ";
            }
            pn = pn + "rmi/"+st.nextToken().trim();
            idx++;
        }
        return pn;            
    }

    /**
     * Stop the server without stopping the JVM
     */
    public void stop() throws Exception {
        if (adm != null) {
            System.out.println("The JOnAS Server '"+serverName+"' has been stopped");
            logger.log(BasicLevel.INFO, "Jonas server "+serverName+" stopped");
            adm.stopServer();
        }
    }

    /**
     * Stop the server and stop the JVM (MBean method)
     */
    public void kill() throws Exception {
        if (adm != null) {
            adm.killServer();
        }
    }


    /**
     * Server main routine.
     * No args.
     */
    public static void main(String args[]) {
        Server server = null;
        try {
            server = Server.getInstance();
            server.start();
        }  catch (JonasAlreadyStartedException e) {
            // if we get this exception this means that
            // we are trying to restart the same JOnAS
            System.err.println("JOnAS error: "+SERVERNAME+" already started");
            System.exit(2);
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: "+e.getMessage());
            System.err.println(SERVERNAME+" halting");
            try {
                if (server!=null) server.stop();
            } catch (Exception ignore) {
            }
            System.exit(2);
        }
    }

}
