/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Michel-Ange ANTON
 * --------------------------------------------------------------------------
 * $Id:
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.webapp.jonasadmin;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.management.ObjectName;

// Temporary
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.jmx.JonasManagementRepr;
import org.objectweb.jonas.jmx.JonasObjectName;

public class JonasAdminJmx {

// ----------------------------------------------------------- Constants

   private static Object s_Synchro = new Object();

// ----------------------------------------------------------- Constructors

   /**
    * Protected constructor to prevent instantiation.
    */
   protected JonasAdminJmx() {
   }

// --------------------------------------------------------- Public Methods

   /**
    * Replace any occurrence of the specified placeholder in the specified
    * template string with the specified replacement value.
    *
    * @param template Pattern string possibly containing the placeholder
    * @param placeholder Placeholder expression to be replaced
    * @param value Replacement value for the placeholder
    * @return A complete string
    */
   public static String replace(String template, String placeholder, String value) {
      if (template == null) {
         return (null);
      }
      if ((placeholder == null) || (value == null)) {
         return (template);
      }
      String sRet = new String(template);
      while (true) {
         int index = sRet.indexOf(placeholder);
         if (index < 0) {
            break;
         }
         StringBuffer temp = new StringBuffer(sRet.substring(0, index));
         temp.append(value);
         temp.append(sRet.substring(index + placeholder.length()));
         sRet = temp.toString();
      }
      return (sRet);

   }

   /**
    * Return a list of JonasServerItem connected to the same registry.
    *
    * @return List of JonasServerItem
    * @throws Exception
    */
   public static ArrayList getListRemoteJonasServerItem()
      throws Exception {
      synchronized (s_Synchro) {
         ArrayList oServers = new ArrayList();
         // Save current RmiConnector
         String sCurrentRMIConnectorName = JonasManagementRepr.getCurrentRMIConnectorName();

         // Find all RmiConnectors
         Set setRmiConnector = JonasManagementRepr.getRMIConnectorsNames();

         if (setRmiConnector.size() > 0) {
            Iterator oIt = setRmiConnector.iterator();
            while (oIt.hasNext()) {
               JonasServerItem oItem = new JonasServerItem();
               oItem.setNameRmiConnector(oIt.next().toString());
               // Find Jonas server name
               JonasManagementRepr.setCurrentRMIConnectorName(oItem.getNameRmiConnector());
               oItem.setNameServer(getJonasServerName());
               oServers.add(oItem);
            }
            try {
               // Restore current RmiConnector
               JonasManagementRepr.setCurrentRMIConnectorName(sCurrentRMIConnectorName);
            }
            catch (Exception e) {
               // May be the current server is shutdown
               if (oServers.size() > 0) {
                  JonasServerItem oItem = (JonasServerItem) oServers.get(0);
                  JonasManagementRepr.setCurrentRMIConnectorName(oItem.getNameRmiConnector());
               }
            }
         }
         return oServers;
      }
   }

   /**
    * Return the current Jonas server name.
    *
    * @return The name
    * @throws Exception
    */
   public static String getJonasServerName()
      throws Exception {
      synchronized (s_Synchro) {
         ObjectName onSearch = JonasObjectName.ejbServer();
         return (String) (JonasManagementRepr.getAttribute(onSearch, "JonasName"));
      }
   }

   /**
    * Verify if the Mbean gotten by the query in the current MbeanServer exists.
    *
    * @param p_On Query Mbean name to search
    * @return true if MBean exists
    * @throws Exception
    */
   public static boolean hasMBeanName(ObjectName p_On)
      throws Exception {
      synchronized (s_Synchro) {
         ArrayList al = new ArrayList();
         Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
         if (itNames.hasNext()) {
            return true;
         }
         return false;
      }
   }

   /**
    * Return the first Mbean name gotten by the query in the current MbeanServer.
    *
    * @param p_On Query Mbean name to search
    * @return The first MBean name or null if not found
    * @throws Exception
    */
   public static String getFirstMBeanName(ObjectName p_On)
      throws Exception {
      synchronized (s_Synchro) {
         ArrayList al = new ArrayList();
         Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
         if (itNames.hasNext()) {
            return itNames.next().toString();
         }
         return null;
      }
   }

   /**
    * Return the list of Mbean name gotten by the query in the current MbeanServer.
    *
    * @param p_On Query Mbean name to search
    * @return The list or null if not found
    * @throws Exception
    */
   public static List getListMBeanName(ObjectName p_On)
      throws Exception {
      synchronized (s_Synchro) {
         ArrayList al = new ArrayList();
         Iterator itNames = JonasManagementRepr.queryNames(p_On).iterator();
         while (itNames.hasNext()) {
            al.add(itNames.next().toString());
         }
         Collections.sort(al);
         return al;
      }
   }

   /**
    * Extract a value of a Mbean name.
    *
    * @param p_Name Name of property
    * @param p_MBeanName Complete Mbean name
    * @return The value or null if not found
    */
   public static String extractValueMbeanName(String p_Name, String p_MBeanName) {
      String sValue = null;
      try {
         String sSearch = p_Name.trim() + "=";
         int iPos = p_MBeanName.indexOf(sSearch);
         if (iPos > -1) {
            sValue = p_MBeanName.substring(iPos + sSearch.length());
            iPos = sValue.indexOf(",");
            if (iPos > -1) {
               sValue = sValue.substring(0, iPos);
            }
         }
      }
      catch (NullPointerException e) {
         // none
         sValue = null;
      }
      return sValue;
   }

   /**
    * Extract the filename of complete path.
    *
    * @param p_Path Complete path (directory and filename)
    * @return The filename or null
    */
   public static String extractFilename(String p_Path) {
      String sFilename = null;
      try {
         int iPosSeparator = p_Path.lastIndexOf("/");
         if (iPosSeparator < 0) {
            iPosSeparator = p_Path.lastIndexOf("\\");
            if (iPosSeparator < 0) {
               sFilename = new String(p_Path);
            }
            else {
               sFilename = p_Path.substring(iPosSeparator + 1);
            }
         }
         else {
            sFilename = p_Path.substring(iPosSeparator + 1);
         }
         if (sFilename.length() > 0) {
            int iPos_1 = p_Path.indexOf(DIR_AUTOLOAD + "/" + sFilename);
            int iPos_2 = p_Path.indexOf(DIR_AUTOLOAD + "\\" + sFilename);
            if (iPos_1 > -1) {
               sFilename = DIR_AUTOLOAD + "/" + sFilename;
            }
            else if (iPos_2 > -1) {
               sFilename = DIR_AUTOLOAD + "\\" + sFilename;
            }
         }
         else {
            sFilename = null;
         }
      }
      catch (NullPointerException e) {
         // none action
         sFilename = null;
      }
      return sFilename;
   }

   /**
    * Temporary declaration of JOnAS directories to wait a direct MBean access.
    */
   private static String DIR_EJBJARS = "ejbjars";
   private static String DIR_APPS = "apps";
   private static String DIR_WEBAPPS = "webapps";
   private static String DIR_RARS = "rars";
   private static String DIR_AUTOLOAD = "autoload";

   private static ArrayList getFilenames(String p_Directory, String p_Extension) {
      ArrayList al = new ArrayList();
      String sExt = "." + p_Extension.toLowerCase();
      String sFile;

      File oFile = new File(p_Directory);
      String[] asFiles = oFile.list();

      if (asFiles != null) {
         for (int i = 0; i < asFiles.length; i++) {
            oFile = new File(p_Directory, asFiles[i]);
            if (oFile.isFile() == true) {
               sFile = oFile.getName().toLowerCase();
               if (sFile.lastIndexOf(sExt) == (sFile.length() - sExt.length())) {
                  al.add(oFile.getName());
               }
            }
         }
      }
      Collections.sort(al);
      return al;
   }

   private static ArrayList getDirectories(String p_Directory) {
      ArrayList al = new ArrayList();
      String sFile;

      File oFile = new File(p_Directory);
      String[] asFiles = oFile.list();

      if (asFiles != null) {
         for (int i = 0; i < asFiles.length; i++) {
            oFile = new File(p_Directory, asFiles[i]);
            if (oFile.isDirectory() == true) {
               al.add(oFile.getName());
            }
         }
      }
      Collections.sort(al);
      return al;
   }

   private static void appendDirectory(ArrayList p_List, String p_Dir) {
      String sDir = p_Dir + "/";
      for (int i = 0; i < p_List.size(); i++) {
         p_List.set(i, sDir + p_List.get(i));
      }
   }

   /**
    * Return the list of JAR filename ready to deploy in the current server.
    *
    * @return The list of JAR filename.
    * @throws Exception
    */
   public static ArrayList getJarFilesDeployable()
      throws Exception {
      ArrayList al = getFilenames(JProp.getJonasBase() + File.separator + DIR_EJBJARS, "jar");
      ArrayList alAuto = getFilenames(JProp.getJonasBase() + File.separator + DIR_EJBJARS
         + File.separator + DIR_AUTOLOAD, "jar");
      appendDirectory(alAuto, DIR_AUTOLOAD);
      al.addAll(alAuto);
      return al;
   }

   /**
    * Return the list of RAR filename ready to deploy in the current server.
    *
    * @return The list of RAR filename.
    * @throws Exception
    */
   public static ArrayList getRarFilesDeployable()
      throws Exception {
      ArrayList al = getFilenames(JProp.getJonasBase() + File.separator + DIR_RARS, "rar");
      ArrayList alAuto = getFilenames(JProp.getJonasBase() + File.separator + DIR_RARS
         + File.separator + DIR_AUTOLOAD, "rar");
      appendDirectory(alAuto, DIR_AUTOLOAD);
      al.addAll(alAuto);
      return al;
   }

   /**
    * Return the list of EAR filename ready to deploy in the current server.
    *
    * @return The list of EAR filename.
    * @throws Exception
    */
   public static ArrayList getEarFilesDeployable()
      throws Exception {
      ArrayList al = getFilenames(JProp.getJonasBase() + File.separator + DIR_APPS, "ear");
      ArrayList alAuto = getFilenames(JProp.getJonasBase() + File.separator + DIR_APPS
         + File.separator + DIR_AUTOLOAD, "ear");
      appendDirectory(alAuto, DIR_AUTOLOAD);
      al.addAll(alAuto);
      return al;
   }

   /**
    * Return the list of WAR filename ready to deploy in the current server.
    *
    * @return The list of WAR filename.
    * @throws Exception
    */
   public static ArrayList getWarFilesDeployable()
      throws Exception {
      ArrayList al = getFilenames(JProp.getJonasBase() + File.separator + DIR_WEBAPPS, "war");
      ArrayList alAuto = getFilenames(JProp.getJonasBase() + File.separator + DIR_WEBAPPS
         + File.separator + DIR_AUTOLOAD, "war");
      appendDirectory(alAuto, DIR_AUTOLOAD);
      al.addAll(alAuto);
      return al;
   }

   /**
    * Return the list of deployed JAR filename in the current server.
    *
    * @return The list of JAR filename.
    * @throws Exception
    */
   public static ArrayList getJarFilesDeployed()
      throws Exception {
      ArrayList alDeployed = new ArrayList();
      String sPath;
      String sFile;
      Iterator itNames = JonasAdminJmx.getListMBeanName(JonasObjectName.allContainers()).iterator();
      while (itNames.hasNext()) {
         sPath = JonasAdminJmx.extractValueMbeanName("fname", itNames.next().toString());
         sFile = JonasAdminJmx.extractFilename(sPath);
         if (sFile != null) {
            alDeployed.add(sFile);
         }
      }
      Collections.sort(alDeployed);
      return alDeployed;
   }

   /**
    * Return the list of deployed EAR filename in the current server.
    *
    * @return The list of EAR filename.
    * @throws Exception
    */
   public static ArrayList getEarFilesDeployed()
      throws Exception {
      ArrayList alDeployed = new ArrayList();
      String sPath;
      String sFile;

      Iterator itNames = ((java.util.Set) JonasManagementRepr.getAttribute(JonasObjectName.
         earService(), "EarNames")).iterator();
      while (itNames.hasNext()) {
         sPath = itNames.next().toString();
         sFile = JonasAdminJmx.extractFilename(sPath);
         if (sFile != null) {
            alDeployed.add(sFile);
         }
      }
      Collections.sort(alDeployed);
      return alDeployed;
   }

   /**
    * Return the list of deployed RAR filename in the current server.
    *
    * @return The list of RAR filename.
    * @throws Exception
    */
   public static ArrayList getRarFilesDeployed()
      throws Exception {
      ArrayList alDeployed = new ArrayList();
      String sPath;
      String sFile;
      Iterator itNames = JonasAdminJmx.getListMBeanName(JonasObjectName.allResourceAdaptors()).
         iterator();
      while (itNames.hasNext()) {
         sPath = JonasAdminJmx.extractValueMbeanName("fname", itNames.next().toString());
         sFile = JonasAdminJmx.extractFilename(sPath);
         if (sFile != null) {
            alDeployed.add(sFile);
         }
      }
      Collections.sort(alDeployed);
      return alDeployed;
   }

   /**
    * Return the list of deployed WAR filename in the current server.
    *
    * @return The list of WAR filename.
    * @throws Exception
    */
   public static ArrayList getWarFilesDeployed()
      throws Exception {
      ArrayList alDeployed = new ArrayList();
      String sPath;
      String sFile;

      Iterator itNames = ((java.util.Set) JonasManagementRepr.getAttribute(JonasObjectName.
         webContainerService(), "WarNames")).iterator();
      while (itNames.hasNext()) {
         sPath = itNames.next().toString();
         sFile = JonasAdminJmx.extractFilename(sPath);
         if (sFile != null) {
            alDeployed.add(sFile);
         }
      }
      Collections.sort(alDeployed);
      return alDeployed;
   }
}