/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import java.util.HashMap;
import java.util.Iterator;

import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.EjbJar;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbJar;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.JonasSession;
import org.objectweb.jonas_ejb.deployment.xml.JonasSessionImpl;
import org.objectweb.jonas_ejb.deployment.xml.Session;
import org.objectweb.util.monolog.api.Logger;


/**
 * Class to hold meta-information related to the deployment of an ejb-jar
 * Contributor(s):
 * Markus Karg: Novell port.
 * Phlippe Durieux: EJB Local References
 * Florent Benoit & Ludovic Bert : Ear
 * @author Christophe Ney [cney@batisseurs.com]
 */
public abstract class DeploymentDesc {

    protected Logger logger;

    /**
     * the role names in the EAR case
     */
    protected String[] earRoleNames = null;

    protected HashMap beanDesc = new HashMap();

    // assembly descriptor
    protected AssemblyDescriptor asd = null;

    protected String displayName = null;


    /**
     * Build the Meta-Information from the XML data binding trees
     * containing the EJB and JOnAS deployment descriptors.
     * @param classLoader The Class Loader to be used
     * @param ejbJar The EjbJar information, from standard deployment descriptor.
     * @param jonasEjbJar The JonasEjbJar information, from JOnAS specific deployment descriptor.
     * @param l The logger to be used for tracing
     * @throws DeploymentDescException Cannot deploy bean
     */
    public DeploymentDesc(ClassLoader classLoader, EjbJar ejbJar, JonasEjbJar jonasEjbJar, Logger l)
        throws DeploymentDescException {
        logger = l;

        // test classloader
        if (classLoader == null) {
            throw new DeploymentDescException("DeploymentDesc: Classloader is null");
        }

        // test the validity of the ejbJar (EnterpriseBeans must be present)
        if (ejbJar.getEnterpriseBeans() == null) {
            throw new DeploymentDescException("DeploymentDesc: Invalid standard deployment descriptor (<enterprise-beans> element missing)");
        }

        // display name
        displayName = (ejbJar.getDisplayName() != null) ? ejbJar.getDisplayName().getValue() : null;

        // assembly descriptor
        asd = ejbJar.getAssemblyDescriptor();

        // HashMap of jonas-session
        HashMap jonasSession = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasSessionList().iterator(); i.hasNext(); ) {
            JonasSession jSes = (JonasSession) i.next();
            jonasSession.put(jSes.getEjbName().getValue(), jSes);
        }

        // session beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getSessionList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            Session ses = (Session) i.next();
            // find corresponding jonas session
            JonasSession jSes = (JonasSession) jonasSession.get(ses.getEjbName().getValue());
            if (jSes == null) {
                // Build a default jonas-session if not exist
                jSes = new JonasSessionImpl();
                jSes.setEjbName(ses.getEjbName());
            }
            if (ses.getSessionType().getValue().equals("Stateful")) {
                // stateful
                bd = new SessionStatefulDesc(classLoader, ses, asd, jSes);
            } else if (ses.getSessionType().getValue().equals("Stateless")) {
                // stateless
                bd = new SessionStatelessDesc(classLoader, ses, asd, jSes);
            } else {
                throw new DeploymentDescException("Invalid session-type content for bean " + ses.getEjbName().getValue());
            }
            bd.setDeploymentDesc(this);
            bd.check();
            beanDesc.put(bd.getEjbName(), bd);
        }

        // HashMap of jonas-entity
        HashMap jonasEntity = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasEntityList().iterator(); i.hasNext(); ) {
            JonasEntity jEnt = (JonasEntity) i.next();
            jonasEntity.put(jEnt.getEjbName().getValue(), jEnt);
        }
        // entity beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getEntityList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            Entity ent = (Entity) i.next();
            // find corresponding jonas entity
            JonasEntity jEnt = (JonasEntity) jonasEntity.get(ent.getEjbName().getValue());
            if (jEnt == null) {
                throw new DeploymentDescException("jonas-entity missing for bean " + ent.getEjbName().getValue());
            }
            if (ent.getPersistenceType().getValue().equals("Bean")) {
                // bean managed
                bd = new EntityBmpDesc(classLoader, ent, asd, jEnt);
            } else if (ent.getPersistenceType().getValue().equals("Container")) {
                // container managed (always jdbc)
                bd = newEntityBeanDesc(classLoader, ent, asd, jEnt);
            } else {
                throw new DeploymentDescException("Invalid persistence-type content for bean " + ent.getEjbName());
            }
            bd.setDeploymentDesc(this);
            bd.check();
            beanDesc.put(bd.getEjbName(), bd);
        }
    }

    /**
     * Get an Iterator on the Bean Desc list
     * @return Iterator on BeanDesc
     */
    public Iterator getBeanDescIterator() {
        return beanDesc.values().iterator();
    }

    /**
     * Get descriptors for all beans contained in jar file
     * @return Array of bean's descriptors
     */
    public BeanDesc[] getBeanDesc() {
        BeanDesc ret[] = new BeanDesc[beanDesc.size()];
        int j = 0;
        for (Iterator i = beanDesc.values().iterator(); i.hasNext(); j++) {
            ret[j] = (BeanDesc) i.next();
        }
        return ret;
    }

    /**
     * Get bean descriptor given its name
     * @param ejbName the name of the bean in the Deployment Descriptor
     */
    public BeanDesc getBeanDesc(String ejbName) {
        return (BeanDesc) beanDesc.get(ejbName);
    }

    /**
     * Get bean descriptor given its abstract schema name
     * @param asn Abstract Schema Name
     * @return null if it doesn't exist.
     */
    public EntityCmp2Desc asn2BeanDesc(String asn) {
        for (Iterator i = beanDesc.values().iterator(); i.hasNext();) {
            BeanDesc bd = (BeanDesc) i.next();
            if (bd instanceof EntityCmp2Desc) {
                if (asn.equals(((EntityCmp2Desc) bd).getAbstractSchemaName())) {
                    return ((EntityCmp2Desc) bd);
                }
            }
        }
        return null;
    }

    /**
     * Get bean descriptor given its interface local name
     * @return null if it doesn't exist.
     */
    public BeanDesc getBeanDescWithLocalInterface(String itfLocalName) {
        for (Iterator i = beanDesc.values().iterator(); i.hasNext();) {
            BeanDesc bd = (BeanDesc) i.next();
            if (bd.getLocalClass() != null) {
                if (itfLocalName.equals(bd.getLocalClass().getName())) {
                    return bd;
                }
            }
        }
        return null;
    }

    protected abstract BeanDesc newEntityBeanDesc(ClassLoader cl, Entity ent, AssemblyDescriptor asd, JonasEntity j) 
        throws DeploymentDescException;

    /**
     * Get the display name
     * @return the Display name string, from the deployment descriptor.
     */
    public String getDisplayName() {
        return displayName;
    }

    /**
     * Get the role names of the deployment desc in the EAR case
     * @return an array of role name
     */
    public String[] getEarRoleNames() {
        return earRoleNames;
    }

    /**
     * Set the role names of the deployment desc in the EAR case
     * @param earRoleNames the role names
     */
    public void setEarRoleNames(String[] earRoleNames) {
        this.earRoleNames = earRoleNames;
    }

    /**
     * get the current logger
     * @return the Logger
     */
    public Logger getLogger() {
        return logger;
    }

    /**
     * set the current logger
     * @param the Logger
     */
    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        BeanDesc[] b = getBeanDesc();
        for (int i = 0; i < b.length; i++) {
            ret.append("\ngetBeanDesc(" + i + ")=" + b[i].getClass().getName());
            ret.append(b[i].toString());
        }
        return ret.toString();
    }

}

