/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): France Telecom R&D, Bull
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.Query;
import org.objectweb.jonas_ejb.deployment.ejbql.ParseException;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jorm.metainfo.api.MetaObject;


/**
 * Class to hold meta-information related to an entity of type CMP version 2.
 * Created on Jun 24, 2002
 * @author Christophe Ney [cney@batisseurs.com]
 * contributor(s):
 * - Helene Joanin on May 2003: code cleanup
 * - Helene Joanin on May 2003: complement for legacy first version 
 */
public abstract class EntityCmp2Desc extends EntityCmpDesc {

    private ArrayList ejbRelationshipRoleDesc = new ArrayList();
    private ArrayList jormMOList = new ArrayList();
    protected String abstractSchemaName;

    private static final String JORM_PACKAGE = "jorm";

    private String jormClassName = null;
    private String jormFQClassName = null;
    private String jormAccessorClassName = null;
    private String jormPKClassName = null;
    private String jormPNameGetterClassName = null;
    private String jormBinderClassName = null;
    private Boolean needCoherenceHelper = null;


    protected String factoryClassName = null;
    protected DeploymentCmp2Desc dc2d = null;

    /**
     * constructor: called when the DeploymentDescriptor is read.
     * Currently, called by both GenIC and createContainer.
     */
    public EntityCmp2Desc(ClassLoader classLoader,
                          Entity ent,
                          AssemblyDescriptor asd,
                          JonasEntity jEnt,
                          DeploymentCmp2Desc dc2d)
        throws DeploymentDescException {

        super(classLoader, ent, asd, jEnt);
        this.dc2d = dc2d;

        // An abstract schema name is required
        if ((ent.getAbstractSchemaName() == null)
            || (ent.getAbstractSchemaName().getValue().length() == 0)) {
            throw new DeploymentDescException("abstract-schema-name must be provided for bean " + this.ejbName);
        }
        abstractSchemaName = ent.getAbstractSchemaName().getValue();

        // check if persistent fields map to getter and setter
        for (Iterator i = fieldDesc.keySet().iterator(); i.hasNext();) {
            String fn = (String) i.next();

            // field should not be defined
            try {
                Field f = ejbClass.getField(fn);
                throw new DeploymentDescException("In cmp-version 2.x, field-name " + fn + " should not be defined in bean " + this.ejbName);
            } catch (NoSuchFieldException e) {
                // this is what we expect: nothing to do
            } catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on " + this.ejbClass.getName());
            }

            try {
                Method getter = null;
                try {
                    getter = ejbClass.getMethod(FieldDesc.getGetterName(fn), null);
                    ((FieldDesc) (fieldDesc.get(fn))).setFieldType(getter.getReturnType());
                } catch (NoSuchMethodException e) {
                    throw new DeploymentDescException("Getter method not found for field-name " + fn + " in bean " + this.ejbName, e);
                }
                try {
                    ejbClass.getMethod(FieldDesc.getSetterName(fn), new Class[]{getter.getReturnType()});
                } catch (NoSuchMethodException e) {
                    throw new DeploymentDescException("Setter method not found for field-name " + fn + " in bean " + this.ejbName, e);
                }
            } catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on " + this.ejbClass.getName());
            }
        }

        // isModifiedMethod deprecated for CMP 2.x
        if (jEnt.getIsModifiedMethodName() != null) {
            throw new DeploymentDescException("use of is-modified-method-name deprecated for CMP 2.x");
        }

        // EJB-QL query
        if (ent.getQueryList() != null) {
            for (Iterator i = ent.getQueryList().iterator(); i.hasNext();) {
                Query q = (Query) i.next();
                boolean foundMatch = false;
                for (Iterator j = getMethodDescIterator(); j.hasNext();) {
                    MethodDesc methd = (MethodDesc) j.next();
                    String methName = q.getQueryMethod().getMethodName().getValue();
                    if (methd.matchPattern(null, methName, q.getQueryMethod().getMethodParams())
                        != MethodDesc.APPLY_TO_NOTHING) {
                        foundMatch = true;
                        String query = q.getEjbQl().getValue();
                        if (!(methd instanceof MethodCmp2Desc)) {
                            throw new DeploymentDescException("ejbql query " + query + " can't apply to method "
                                                              + methName + " in bean " + ejbName);
                        }
                        try {
                            ((MethodCmp2Desc) methd).setQuery(query);
                        } catch (ParseException e) {
                            throw new DeploymentDescException("Invalid ejbql syntax for bean " + ejbName + ":\n"
                                                              + e.getMessage(query));
                        }
                        if (q.getResultTypeMapping() != null) {
                            ((MethodCmp2Desc) methd).setResultTypeMapping(q.getResultTypeMapping().getValue());
                        }
                    }
                }
                if (!foundMatch) {
                    throw new DeploymentDescException("invalid query-method definition for bean" + ejbName + "\nno such method as "
                                                      + MethodCmp2Desc.queryMethodElementToString(q.getQueryMethod()) + "\ncheck method name and method parameters");
                }
            }
        }
        // check that all finder/selecter methods but findByPrimaryKey have a non null query
        for (Iterator j = getMethodDescIterator(); j.hasNext();) {
            MethodDesc md = (MethodDesc) j.next();
            if ((md.isFinder() || md.isEjbSelect()) && !md.isFindByPrimaryKey()) {
                if (((MethodCmp2Desc) md).getQuery() == null) {
                    throw new DeploymentDescException("query not defined for method " + MethodDesc.toString(md.getMethod())
                                                      + " of bean" + ejbName);

                }
            }
        }

    }

    public DeploymentCmp2Desc getDeploymentCMP2Desc() {
        return dc2d;
    }

    /**
     * getter for field abstractSchemaName
     */
    public String getAbstractSchemaName() {
        return abstractSchemaName;
    }

    /**
     * Get the Jorm Class name in JOrm Meta Info
     * It is built from the Abstract Shema Name.
     * @return the Jorm Class Name
     */
    private String getJormCName() {
        if (jormClassName == null) {
            jormClassName = BeanNaming.firstToUpperCase(abstractSchemaName);
        }
        return jormClassName;
    }

    /**
     * Get the Jorm Fully Qualified Class name in JOrm Meta Info
     * It is built from the Abstract Shema Name.
     * @return the Jorm Class Name
     */
    public String getJormClassName() {
        if (jormFQClassName == null) {
            jormFQClassName = BeanNaming.getClassName(JORM_PACKAGE, getJormCName());
        }
        return jormFQClassName;
    }

    /**
     * Get the list of Jorm MetaObjects to be generated
     * @return a Collection of Jorm Meta Objects
     */
    public Collection getJormList() {
        return jormMOList;
    }

    /**
     * add a Jorm MetaObject in the list of Jorm Meta Objects to generate
     * @param jobject the Jorm MetaObject to add in the list
     */
    public void addToJormList(MetaObject jobject) {
        jormMOList.add(jobject);
    }

    /**
     * Add meta-info of a relationship in which the bean is involved
     */
    public void addEjbRelationshipRoleDesc(EjbRelationshipRoleDesc ersrd) {
        ejbRelationshipRoleDesc.add(ersrd);
    }

    /**
     * Get Iterator on meta-info for all relationships for which the bean
     * is involved in.
     */
    public Iterator getEjbRelationshipRoleDescIterator() {
        return ejbRelationshipRoleDesc.iterator();
    }

    /**
     * Get the EjbRelationshipRoleDesc corresponding to the given cmr field name.
     * Return null if it doesn't exist.
     */
    public EjbRelationshipRoleDesc getEjbRelationshipRoleDesc(String cmr) {
        for (Iterator i = ejbRelationshipRoleDesc.iterator(); i.hasNext();) {
            EjbRelationshipRoleDesc rsr = (EjbRelationshipRoleDesc) i.next();
            if (rsr.hasCmrField() && cmr.equals(rsr.getCmrFieldName())) {
                return rsr;
            }
        }
        return null;
    }

    /**
     * get iterator on meta information for all cmp fields
     */
    public Iterator getCmpFieldDescIterator() {
        return fieldDesc.values().iterator();
    }

    /**
     * Factory method for MethodDesc.
     * Only method with EJBQL queries are of type <code>MethodCmp2Desc</code>
     */
    protected MethodDesc newMethodDescInstance(Method meth, int index) {
        return new MethodCmp2Desc(this, meth, index);
    }

    /**
     * It retrieves the class name of the generated Jorm accessor interface.
     * Used in the templates
     * @return the fully qualified class name
     */
    public String getJormAccessorClassName() {
        if (jormAccessorClassName == null) {
            jormAccessorClassName = BeanNaming.getClassName(JORM_PACKAGE, getJormCName() + "Accessor");
        }
        return jormAccessorClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm binding.
     * Used by the JormFactory
     * @param mapper The mapper name
     * @return the fully qualified class name
     */
    public String getJormBindingClassName(String mapper) {
        return BeanNaming.getClassName(JORM_PACKAGE, mapper + "." + getJormCName() + "Binding");
    }

    /**
     * It retrieves the class name of the generated Jorm mapping.
     * Used in the JContainer to instanciate the JEntityFactory
     * @param mapper The mapper name
     * @return the fully qualified class name
     */
    public String getFactoryClassName(String mapper) {
        return BeanNaming.getClassName(JORM_PACKAGE, mapper + "." + getJormCName() + "Mapping");
    }

    /**
     * Retrieves the Jorm PK Class, in case of composite PK
     * Used by Jorm to generate PNameGetter and Binder classes
     * @return the fully qualified class name
     */
    public String getJormPKClassName() {
        if (! hasPrimaryKeyField() && jormPKClassName == null) {
            jormPKClassName = BeanNaming.getClassName(JORM_PACKAGE, 
                                                      BeanNaming.getBaseName(getPrimaryKeyClass().getName()));
        }
        return jormPKClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm PNameGetter interface.
     * Used in the templates to generate CMP2 classes.
     * @return the fully qualified class name
     */
    public String getJormPNameGetterClassName() {
        if (! hasPrimaryKeyField() && jormPNameGetterClassName == null) {
            jormPNameGetterClassName = getJormPKClassName() + "PNG";
        }
        return jormPNameGetterClassName;
    }

    /**
     * It retrieves the class name of the generated Jorm Binder.
     * Used in the Jorm Factory
     * @return the fully qualified class name
     */
    public String getJormBinderClassName() {
        if (jormBinderClassName == null) {
            if (hasPrimaryKeyField()) {
                jormBinderClassName = "org.objectweb.jorm.facility.naming.basidir.BasidBinder";
            } else {
                jormBinderClassName = getJormPKClassName() + "Binder";
            }
        }
        return jormBinderClassName;
    }

    // TODO: boolean as return value instead of java.lang.Boolean
    public Boolean needJormCoherenceHelper() {
        if (needCoherenceHelper == null) {
            needCoherenceHelper = (ejbRelationshipRoleDesc.iterator().hasNext()
                                   ? Boolean.TRUE : Boolean.FALSE);
        }
        return needCoherenceHelper;
    }

    public String getJormCoherenceHelperItfName() {
        return "JOnAS" + ejbName + "CoherenceHelper";
    }

    public String getJormCoherenceHelperPackageName() {
        return BeanNaming.getPackageName(getFullDerivedBeanName());
    }

    public String getJormCoherenceHelperFQItfName() {
        String pn = getJormCoherenceHelperPackageName();
        return (pn != null && pn.length() > 0
                ? pn + "." + getJormCoherenceHelperItfName()
                : getJormCoherenceHelperItfName());
    }

    /**
     * String representation of the object for test and debug purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        for (Iterator i = ejbRelationshipRoleDesc.iterator(); i.hasNext(); ) {
            ret.append("\nejbRelationshipRoleDesc[]=" + i.next());
        }
        ret.append("\ngetAbstractSchemaName()=" + getAbstractSchemaName());
        ret.append("\ngetJormAccessorClassName() = " + getJormAccessorClassName());
        ret.append("\nneedJormCoherenceHelper() = " + needJormCoherenceHelper());
        return ret.toString();
    }

}

