/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Tecnologies Inc. , France Telecom R&D
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.deployment.api;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Iterator;

import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.CmpField;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.lib.BeanNaming;

/**
 * Base class to hold meta-information related to an entity of type CMP.
 * @author Christophe Ney [cney@batisseurs.com]
 */
public abstract class EntityCmpDesc extends EntityDesc {

    protected HashMap fieldDesc = new HashMap();
    protected String primaryKeyField = null;

    /**
     * constructor to be used by parent node.
     * @param classLoader class loader to use to laod bean's classes
     * @param ent entity in the standard deployment descriptor
     * @param asd assembly-descriptor in the standard deployment descriptor
     * @param jEnt jonas-entity in the JOnAS deployment descriptor
     * @throws DeploymentDescException in error case.
     */
    public EntityCmpDesc(ClassLoader classLoader,
                         Entity ent, AssemblyDescriptor asd, JonasEntity jEnt)
        throws DeploymentDescException {
        super(classLoader, ent, asd, jEnt);

        // field descriptors for persistent fields
        for (Iterator i = ent.getCmpFieldList().iterator(); i.hasNext();) {
            String fn = ((CmpField) i.next()).getFieldName().getValue();
            FieldDesc fd = this.newFieldDescInstance();
            fd.setName(fn);
            fd.setPrimaryKey(false);
            fieldDesc.put(fn, fd);
        }

        if (ent.getPrimkeyField() != null) {
            // primary key field
            primaryKeyField = ent.getPrimkeyField().getValue();
            FieldDesc fd = (FieldDesc) fieldDesc.get(primaryKeyField);
            if (fd == null) {
                throw new DeploymentDescException("primkey-field " + primaryKeyField + " is not listed as cmp-field in bean " + this.ejbName);
            }
            fd.setPrimaryKey(true);
        } else {
            // public fields of primary key class
            Field[] pcf = primaryKeyClass.getFields();
            for (int i = 0; i < pcf.length; i++) {
                if (Modifier.isPublic(pcf[i].getModifiers())) {
                    String pn = pcf[i].getName();
                    // exclude serialVersionUID field for jdk1.2.1 on solaris and
                    // exclude JProbe$ for JProbe using
                    if (!pn.equals("serialVersionUID") && !pn.startsWith("JProbe$")) {
                        FieldDesc fd = (FieldDesc) fieldDesc.get(pn);
                        if (fd == null) {
                            throw new DeploymentDescException("public field " + pn + " of primkey-class is not listed as cmp-field in bean " + this.ejbName);
                        }
                        fd.setPrimaryKey(true);
                    }
                }
            }
        }
        String packageName = BeanNaming.getPackageName(getEjbClass().getName());
        derivedBeanName = new String("JOnAS" + getIdentifierFromEjbName(ejbName) + "Bean");
        fullDerivedBeanName = BeanNaming.getClassName(packageName, derivedBeanName);
    }

    /**
     * Get descriptor for a given field name
     * Used by GenIC
     * @param fieldName Name of the field
     * @return Descriptor for the given field or null
     */
    public FieldDesc getCmpFieldDesc(String fieldName) {
        return (FieldDesc) fieldDesc.get(fieldName);
    }

    /**
     * Indicate if the primary key field is only one field with a primary-key-field
     * defined in the DD.
     * @return true if the primary key field is only one field
     *         with a primary-key-field defined in the DD.
     */
    public boolean hasSimplePkField() {
        return primaryKeyField != null;
    }
    // TODO: remove this method and keep only the hasSimplePkField method
    public boolean hasPrimaryKeyField() {
        return hasSimplePkField();
    }

    /**
     * Get the primary key field for the entity.
     * @return Field for the primary key
     */
    public FieldDesc getSimplePkField() {
        FieldDesc fd = (FieldDesc) fieldDesc.get(getSimplePkFieldName());
        return fd;
    }

    /**
     * Get the primary key field name for the entity.
     * @return Field for the primary key
     */
    public String getSimplePkFieldName() {
        if (primaryKeyField == null) {
            throw new Error("No primary key field defined for bean " + this.ejbName);
        }
        return primaryKeyField;
    }
    // TODO: remove this method and keep only the getSimplePkFieldName() method
    public String getPrimaryKeyFieldName() {
        return getSimplePkFieldName();
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        for (Iterator i = fieldDesc.keySet().iterator(); i.hasNext();) {
            String f = (String) i.next();
            FieldDesc fd = (FieldDesc) fieldDesc.get(f);
            ret.append("\ngetCmpFieldDesc(" + f + ")=" + fd.getClass().getName());
            ret.append(fd.toString());
        }
        if (hasPrimaryKeyField()) {
            ret.append("\ngetPrimaryKeyField()=" + getPrimaryKeyFieldName());
        }
        return ret.toString();
    }

    /**
     * factory method for field descriptors
     * @return a FieldDesc.
     */
    protected FieldDesc newFieldDescInstance() {
        return new FieldJdbcDesc();
    }
}

