/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributor(s):
 *   01/05/30 Christophe Ney - Lutris Technologies
 *          Fix to handle arguments containing white spaces.
 *   01/11/23 Guillaume Riviere
 *          Fix the addClassesInJar() method
 *          (on David, the Stub/Skel classes names are different).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.genic;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.net.URLClassLoader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;

import org.objectweb.common.Cmd;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.Version;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.api.MedorException;


/**
 *  This class allows to generate:
 * <ul>
 * <li>
 *  the classes that implements the Enterprise bean's remote interface,
 * <li>
 *  the classes that implements the Enterprise bean's home interface,
 * <li>
 *  the classes that implements the Handles of the Entity beans,
 * <li>
 *  the classes that implements the persistence of the Entity beans with CMP,
 * </ul>
 * of all the Enterprise Java Beans defined in the given Deployment Descriptor.
 */

public class GenIC {

    public static final String RMI_JRMP = "jrmp";
    public static final String RMI_IIOP = "iiop";
    public static final String JEREMIE  = "jeremie";
    public static final String CMI_RMI = "cmi";

    private static final int EXIT_SUCCESS = 0;
    private static final int EXIT_FAILURE = 2;

    // java.home property value (ended with '/')
    private static String javaHomeBin = null;

    // Is the command is verbose ?
    private boolean verbose = false;

    // Is the command is on debug mode ?
    private boolean debug = false;

    // Name of the directory where to place the generated file
    private String outputdir = null;

    // Are some container classes generated ?
    private boolean generatedIC;

    // Path names of the generated files to delete
    private Vector filesToDelete = null;

    // Path names of the java generated sources (remote/not remote)
    private Vector remoteJavas = null;
    private Vector noRemoteJavas = null;

    // Names of the remote classes (package name included)
    private Vector remoteClasses = null;

    // Path and classes of files generated for clustering
    private ArrayList clusterConfs = null;
    private ArrayList clusteredClasses = null;


    /**
     * This field references the jorm meta info manager, and in particular the
     * manager provided by JOnAS. Indeed it is able to build the jorm meta
     * information from the jonas meta information (deployment/api)
     */
    private Manager jormManager = null;

    private static Logger logger = null;

    private static String getClassPath(URLClassLoader ucl) {
        String ps = System.getProperty("path.separator");
        String cp = ".";
        URL[] urls = ucl.getURLs();
        for (int i = 0; i < urls.length; i++) {
            cp = cp + ps + urls[i].getFile();
        }
        return cp;
    }

    /**
     * GenIC allows to generate the container classes for JOnAS for the given Enterprise Java Beans.
     * <p>
     *
     * Usage: java org.objectweb.jonas_ejb.genic.GenIC -help
     * <br>
     *         to print this help message
     * <p>
     *
     * or    java org.objectweb.jonas_ejb.genic.GenIC <Options> <Input_File>
     * <br>
     *         to generate the container classes for the given EJBs.
     * <p>
     *           Options include:
     * <ul>
     * <li>
     *             -d <output_dir>  specify where to place the generated files
     * </li>
     * <li>
     *             -noaddinjar      do not add the generated classes in the given
     *                              ejb-jar file
     * </li>
     * <li>
     *             -nocompil        do not compile the generated source files via
     *                              the java and rmi compilers
     * </li>
     * <li>
     *             -novalidation    parse the XML deployment descriptors without validation
     * </li>
     * <li>
     *             -javac     <opt> specify the name of the java compiler to use
     * </li>
     * <li>
     *             -javacopts <opt> specify the options to pass to the java compiler
     * </li>
     * <li>
     *             -rmicopts <opt>  specify the options to pass to the rmi compiler
     * </li>
     * <li>
     *             -secpropag       modify the RMI Skel. and Stub. to implement
     *                              the implicit propagation of the security context
     * </li>
     * <li>
     *             -mappernames     list of JORM mapper names, separated by comma
     * </li>
     * <li>
     *             -protocols       list of protocol, separated by comma
     * </li>
     * <li>
     *             -keepgenerated   do not delete intermediate generated source files
     * </li>
     * <li>
     *             -verbose
     * </li>
     * <li>
     *             -debug
     * </li>
     * </ul>
     * <p>
     *           Input_File         file name of the standard deployment descriptor (.xml ended),
     *                              or file name of the EJB-jar (.jar ended).
     */
    
    public static void main(String args[]) {
        boolean  error = false;

        boolean  isHelp          = false;
        boolean  isVerbose       = false;
        boolean  isDebug         = false;
        boolean  isKeepGenerated = false;
        boolean  isSecPropag     = false;
        boolean  isCompil        = true;
        boolean  isAddInJar      = true;
        boolean  parseWithValidation = true;

        String   fileInputName = null;
        String   dirOutputName = null;
        String   nameJavac     = null;
        Vector   optionsJavaC  = new Vector();
        Vector   optionsRmiC   = new Vector();
        String   protocols = RMI_JRMP + "," + JEREMIE; // By default we generate for jrmp & jeremie
        ProtocolNames  protocolsNames = null;
        String   cmiLb         = null;

        // Get args
        for (int argn = 0; argn < args.length; argn++) {
            String arg = args[argn];
            if (arg.equals("-help") || arg.equals("-?")) {
                isHelp = true;
                continue;
            }
            if (arg.equals("-verbose")) {
                isVerbose = true;
                continue;
            }
            if (arg.equals("-debug")) {
                isDebug = true;
                isVerbose = true;
                continue;
            }
            if (arg.equals("-mappernames")) {
                argn++;
                if (argn < args.length) {
                    System.setProperty("JONAS_CMP_MAPPERNAMES", args[argn]);
                    continue;
                } else {
                    error = true;
                }
            }
            if (arg.equals("-protocols")) {
                argn++;
                if (argn < args.length) {
                    protocols =  args[argn];
                    continue;
                } else {
                    error = true;
                }
                
            }
            if (arg.equals("-keepgenerated")) {
                isKeepGenerated = true;
                optionsRmiC.addElement(args[argn]);
                continue;
            }
            if (arg.equals("-nocompil")) {
                isCompil = false;
                isKeepGenerated = true;
                continue;
            }
            if (arg.equals("-noaddinjar")) {
                isAddInJar = false;
                continue;
            }
            if (arg.equals("-novalidation")) {
                parseWithValidation = false;
                continue;
            }
            if (arg.equals("-javac")) {
                argn++;
                continue;
            }
            if (arg.equals("-javacopts")) {
                argn++;
                if (argn < args.length) {
                    StringTokenizer st = new StringTokenizer(args[argn]);
                    while (st.hasMoreTokens()) {
                        optionsJavaC.addElement(st.nextToken());
                    }
                } else {
                    error = true;
                }
                continue;
            }
            if (arg.equals("-rmicopts")) {
                argn++;
                if (argn < args.length) {
                    StringTokenizer st = new StringTokenizer(args[argn]);
                    while (st.hasMoreTokens()) {
                        optionsRmiC.addElement(st.nextToken());
                    }
                } else {
                    error = true;
                }
                continue;
            }
            if (arg.equals("-cmilb")) {
                argn++;
                if ((cmiLb == null) && (argn < args.length)) {
                    cmiLb = args[argn];
                } else {
                    error = true;
                }
                continue;
            }
            if (arg.equals("-d")) {
                argn++;
                if (argn < args.length) {
                    dirOutputName = args[argn];
                } else {
                    error = true;
                }
                continue;
            }
            fileInputName = args[argn];
        }

        Log.configure("trace");

        // Usage ?
        if (isHelp) {
            usage();
            System.exit(EXIT_SUCCESS);
        }

        // Check args
        if (error || (fileInputName == null)) {
            usage();
            System.exit(EXIT_FAILURE);
        }
        // The -d option is deprecated since JOnAS 3.0.7 when the file input is an ejb-jar
        // Instead, the output directory must be a temporary directory to make easier
        // the ejb-jar updated.
        if ((dirOutputName != null) && isAddInJar && fileInputName.endsWith(".jar")) {
            warning("The -d '" + dirOutputName + "' option is ignored"
                    + "(deprecated with an ejb-jar as input file)");
        }
            
        if (dirOutputName == null) {
            dirOutputName = new String("");
        }

        // Build the array of protocols name 
        protocolsNames = new ProtocolNames(protocols);

        // In case of ejb-jar file, the dirOutputName must be initialized with a tempo. directory.
        if (fileInputName.endsWith(".jar") && isAddInJar) {
            try {
                dirOutputName = GenIC.createTempDir();
            } catch (IOException ioe) {
                GenIC.fatalError(ioe);
            }
        }

        // Init the classpath used for javac, rmic, JRMICompiler
        URLClassLoader pcl = (URLClassLoader) Thread.currentThread().getContextClassLoader();
        String classpath = getClassPath(pcl);
        if (fileInputName.endsWith(".jar")) {
            classpath = fileInputName + System.getProperty("path.separator") + classpath;
        }
        if (!"".equals(dirOutputName)) {
            classpath = dirOutputName + System.getProperty("path.separator") + classpath;
        }

        // Set the parsing mode (with or without validation)
        if (!parseWithValidation) {
            EjbDeploymentDescManager.setParsingWithValidation(false);
        }

        // Generates the classes for the set of the beans
        try {
            DeploymentDesc ejbJarDD = null;
            if (fileInputName.endsWith(".jar")) {
                // ejb-jar file
                URL url[] = new URL[1];
                url[0] = (new File(fileInputName)).toURL();
                URLClassLoader cl = new URLClassLoader(url, pcl);
                ejbJarDD = EjbDeploymentDescManager.getDeploymentDesc(fileInputName, cl);
            } else {
                // xml file
                ejbJarDD = EjbDeploymentDescManager.getDeploymentDesc(fileInputName,
                                                                      BeanNaming.getJonasXmlName(fileInputName));
            }

            GenIC gwc = new GenIC(ejbJarDD, dirOutputName, isVerbose, isDebug);

            // Reset the beans deployment descriptors to unload the beans classes
            // (loaded from the ejb-jar when creating the jonas_ejb.deployment.api.BeanDesc),
            // to be able to update the ejb-jar file on Windows.
            // See Bug #270
            ejbJarDD = null;
            System.gc();

            if (isCompil) {
                gwc.compilClasses(nameJavac, optionsJavaC, optionsRmiC, classpath, protocolsNames, isKeepGenerated);
                if ( protocolsNames.isSupported(CMI_RMI)) {
                    gwc.genClusterFiles(classpath, isKeepGenerated);
                }
                if (fileInputName.endsWith(".jar") && isAddInJar) {
                    gwc.addClassesInJar(fileInputName, isKeepGenerated);
                }
                if (!isKeepGenerated) {
                    gwc.clean();
                }
            }

        } catch (MalformedURLException e) {
            GenIC.fatalError("Invalid ejb-jar file name (" + e.getMessage() + ")");
        } catch (GenICException e) {
            GenIC.fatalError(e);
        } catch (DeploymentDescException e) {
            GenIC.fatalError("Cannot read the Deployment Descriptors from "
                             + fileInputName + ": " + e.getMessage());
        }
        // End
    }

    /**
     * GenIC Constructor: generates the container classes sources of each beans
     *
     * @param     ejbJarDesc     deployment descriptors of the beans
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @param     isVerbose      if true, some traces are print
     * @param     isDebug        if true, many debug traces are print
     * @throws GenICException In error case
     */
    public GenIC(DeploymentDesc ejbJarDesc, String dirOutputName,
                 boolean isVerbose, boolean isDebug) throws GenICException {

        // A BeanSources for each bean
        Vector beanList = null;

        verbose = isVerbose;
        logger = Log.getLogger(Log.JONAS_GENIC_PREFIX);

        if (javaHomeBin == null) {
            javaHomeBin = System.getProperty("java.home", "");
            if (!("".equals(javaHomeBin))) {
                javaHomeBin = javaHomeBin + File.separator + ".."
                    + File.separator + "bin" + File.separator;
            }
        }
        outputdir = dirOutputName;
        filesToDelete = new Vector();
        remoteJavas = new Vector();
        noRemoteJavas = new Vector();
        remoteClasses = new Vector();
        clusterConfs = new ArrayList();
        clusteredClasses = new ArrayList();
        ejbJarDesc.setLogger(logger);
        beanList = new Vector();
        BeanDesc[] beansDD = ejbJarDesc.getBeanDesc();
        for (int i = 0; i < beansDD.length; i++) {
            if (isDebug) {
                info(beansDD[i].toString());
            }
            // Nothing to generate in case of MessageDriven Bean
            if (!(beansDD[i] instanceof MessageDrivenDesc)) {
                // Load the jorm meta information of the class
                if (beansDD[i] instanceof EntityCmp2Desc) {
                    if (jormManager == null) {
                        try {
                            jormManager = ((DeploymentCmp2Desc) ejbJarDesc).getJormManager();
                        } catch (DeploymentDescException e) {
                            throw new GenICException("Impossible to load jorm meta information", e);
                        }
                    }
                    beanList.addElement(new BeanSources(beansDD[i], outputdir, jormManager, logger));
                } else {
                    beanList.addElement(new BeanSources(beansDD[i], outputdir, logger));
                }
            }
        }

        // Display the bean's names
        StringBuffer message = new StringBuffer();
        message.append("GenIC for JOnAS " + Version.NUMBER + ": ");
        if (beanList.size() != 0) {
            generatedIC = true;
            message.append("Beans ");
            for (int i = 0; i < beanList.size(); i++) {
                BeanSources ics = (BeanSources) beanList.elementAt(i);
                message.append((i == 0 ? "" : ", ") + "'" + ics.getEjbName() + "'");
            }
            message.append(" generation ...");
        } else {
            generatedIC = false;
            message.append("No generation to do (only message driven beans)");
        }
        GenIC.info(message.toString());

        // Generates the sources of the container classes of the beans
        // and Init the lists of the remote/non-remote java sources and the remote classes
        for (Enumeration e = beanList.elements(); e.hasMoreElements();) {
            BeanSources ics = (BeanSources) e.nextElement();
            ics.generate();
            String ccfn = ics.getWrpHomeClusterConfFileName();
            if (ccfn != null) {
                filesToDelete.add(ccfn);
            }
            trace("Sources classes successfully generated" + " for '" + ics.getEjbName() + "'");
            noRemoteJavas.addAll(ics.getNoRemoteJavas());
            if (ics.getWrpHomeFileName() != null) {
                remoteJavas.addElement(ics.getWrpHomeFileName());
                remoteClasses.addElement(ics.getWrpHomeClassName());
            }
            if (ics.getWrpRemoteFileName() != null) {
                remoteJavas.addElement(ics.getWrpRemoteFileName());
                remoteClasses.addElement(ics.getWrpRemoteClassName());
            }
            if (ics.getWrpHomeClusterConfFileName() != null) {
                clusterConfs.add(ics.getWrpHomeClusterConfFileName());
                clusteredClasses.add(ics.getWrpHomeClassName());
            }
        }

    }


    /**
     * Compiles the java sources generated by the constructor
     *
     * I.e. :
     * <ul>
     * <li>
     *   compile the classes via javac,
     * <li>
     *   create  the stubs and skeletons for the protocols supported(RMI/JRMP, RMI/IIOP,JEREMIE)
     *   via rmic/jrmic
     * </ul>
     *
     * @param     nameJavac      name the javac tool to use (if null, take the standard tool)
     * @param     optionsJavaC   options for the javac tool
     * @param     optionsRmiC    options for the rmic tool
     * @param     classpath      classpath value
     * @param     protocols      names of protocols 
     * @exception GenICException In error case
     */
    public void compilClasses(String nameJavac,
                              Vector optionsJavaC,
                              Vector optionsRmiC,
                              String classpath,
                              ProtocolNames protocols,
                              boolean keepGenerated)
        throws GenICException {

        String optDirOutput;
        Cmd jrmpCmd;
        Cmd iiopCmd;
        Cmd jeremieCmd;
        Cmd cmd;

        String cmdJava;
        String cmdJavac;
        String cmdRmic;

        if (!generatedIC) {
            return;
        }

        /*
         * Init the command names
         */
        cmdJava = javaHomeBin + "java";
        cmdJavac = (nameJavac == null) ? (javaHomeBin + "javac") : nameJavac;
        cmdRmic = javaHomeBin + "rmic";

        /*
         * Compile the generated sources
         */
        cmd = new Cmd(cmdJavac);
        cmd.addArgument("-classpath");
        cmd.addArgument(classpath);
        cmd.addArgument("-d");
        cmd.addArgument((outputdir.length() == 0) ? "." : outputdir);
        cmd.addArguments(optionsJavaC);
        for (Enumeration e = remoteJavas.elements(); e.hasMoreElements();) {
            String srcName = (String) e.nextElement();
            cmd.addArgument(srcName);
            filesToDelete.addElement(srcName);
        }
        for (Enumeration e = noRemoteJavas.elements(); e.hasMoreElements();) {
            String srcName = (String) e.nextElement();
            cmd.addArgument(srcName);
            filesToDelete.addElement(srcName);
        }

        trace("Running '" + cmd.toString() + "'");
        if (cmd.run()) {
            trace("Sources classes successfully compiled via java compiler.");
        } else {
            throw new GenICException("Failed when compiling the generated classes via java compiler");
        }

        if (remoteJavas.size() == 0) {
            return;
        }

        /*
         * Generate the stub and skeletons of the home and remote implementations
         */
        // compile the rmic jrmp class (version 1.2 and 1.1)
        jrmpCmd = new Cmd(cmdRmic);
        jrmpCmd.addArgument("-classpath");
        jrmpCmd.addArgument(classpath);
        jrmpCmd.addArgument("-keepgenerated");

        // compile the rmic iiop class
        iiopCmd = new Cmd(cmdRmic);
        iiopCmd.addArgument("-classpath");
        iiopCmd.addArgument(classpath);
        iiopCmd.addArgument("-keepgenerated");
        iiopCmd.addArgument("-iiop");
        iiopCmd.addArgument("-always");

        // compile the jeremie stub 
        jeremieCmd = new Cmd(cmdJava);
        jeremieCmd.addArgument("-classpath");
        jeremieCmd.addArgument(classpath);
        jeremieCmd.addArgument("org.objectweb.jeremie.tools.jrmic.JRMICompiler");
        jeremieCmd.addArgument("-opt");
        jeremieCmd.addArgument("-owext");


        if (!"".equals(outputdir)) {
            jrmpCmd.addArgument("-d");
            jrmpCmd.addArgument(outputdir);
            iiopCmd.addArgument("-d");
            iiopCmd.addArgument(outputdir);
            jeremieCmd.addArgument("-d");
            jeremieCmd.addArgument(outputdir);
        }
        jrmpCmd.addArguments(optionsRmiC);
        iiopCmd.addArguments(optionsRmiC);
        jeremieCmd.addArguments(optionsRmiC);
        for (Enumeration e = remoteClasses.elements(); e.hasMoreElements();) {
            String className = (String) e.nextElement();
            jrmpCmd.addArgument(className);
            iiopCmd.addArgument(className);
            jeremieCmd.addArgument(className);
        }

        if (protocols.isSupported(RMI_JRMP)) {
            trace("Running '" + jrmpCmd.toString() + "'");
            if (jrmpCmd.run()) {
                info("Stubs and Skels successfully generated for rmi/jrmp");
            } else {
                throw new GenICException("Failed when generating the Stubs and Skels with rmic jrmp");
            }
        }

        if (protocols.isSupported(RMI_IIOP)) {
            trace("Running '" + iiopCmd.toString() + "'");
            if (iiopCmd.run()) {
                info("Stubs and Skels successfully generated for rmi/iiop");
            } else {
                throw new GenICException("Failed when generating the Stubs and Skels with rmic iiop");
            }

        }

        if (protocols.isSupported(JEREMIE)) {
            trace("Running '" + jeremieCmd.toString() + "'");
            if (jeremieCmd.run()) {
                info("Stubs and Skels successfully generated with rmi/jeremie");
            } else {
                throw new GenICException("Failed when generating the Stubs and Skels with jeremie jrmic");
            }
        }
    }

    private static Class clusterCompilerClass = null;
    private static Method clusterCompilerGenerateMethod = null;

    public static void callClusterCompiler(String[] args) throws ClassNotFoundException, InstantiationException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
        if (clusterCompilerClass == null) {
            clusterCompilerClass = Class.forName("org.objectweb.carol.cmi.compiler.Compiler");
            clusterCompilerGenerateMethod = clusterCompilerClass.getMethod("generate", new Class[] { String[].class });
        }
        Object cc = clusterCompilerClass.newInstance();
        clusterCompilerGenerateMethod.invoke(cc, new Object[] { args });
    }

    /**
     * Generate the cluster stub files from the cluster config files.
     */
    public void genClusterFiles(String classpath, boolean isKeepGenerated) throws GenICException {
        Iterator cf = clusterConfs.iterator();
        Iterator cl = clusteredClasses.iterator();
        while (cf.hasNext()) {
            String ccfn = (String) cf.next();
            String cn = (String) cl.next();
        
            try {
                Vector args = new Vector();
                if (isKeepGenerated) args.addElement("-keep");
                args.addElement("-conf");
                args.addElement(ccfn);
                args.addElement("-classpath");
                args.addElement(classpath);
                if (!"".equals(outputdir)) {
                    args.addElement("-d");
                    args.addElement(outputdir);
                }
                args.addElement(cn);
        
                String[] c = new String[args.size()];
                args.copyInto(c);
                callClusterCompiler(c);
                
            } catch (Exception ex) {
                throw new GenICException(
                    "Failed when generating the cluster stubs",
                    ex);
            }
        }
        info("Cluster stubs successfully generated");
    }

    /**
     * Add the generated classes in the given ejb-jar file.
     */
    public void addClassesInJar(String jarFileName, boolean keepGenerated) throws GenICException {

        if (!generatedIC) {
            return;
        }

        filesToDelete.addElement(outputdir);

        ArrayList lf = new ArrayList();
        getFilesList(outputdir, lf, !keepGenerated);
        // Call the jar command several times to avoid a too long command
        final int NB_MAX_FILES = 20;
        int nbf = lf.size();
        for (int i = 0; i < nbf; i += NB_MAX_FILES) {
            List slf = lf.subList(i, (i + NB_MAX_FILES < nbf) ? i + NB_MAX_FILES : nbf);
            addSomeClassesInJar(jarFileName, outputdir, slf);
        }

    }

    private void addSomeClassesInJar(String jfn, String od, List lfn) throws GenICException {
        /*
         * Example of a jar command:
         *   jar -uf ../output/ejbjars/sb.jar -C /tmp/genic1547.tmp a/b/C1.class
         *                                    -C /tmp/genic1547.tmp a/b/C2.class
         *                                    .....
         */
        Cmd cmd = null;
        cmd = new Cmd(javaHomeBin + "jar");
        cmd.addArgument("-uf");
        cmd.addArgument(jfn);
        int is = od.length() + 1;
        Iterator ilf = lfn.iterator();
        while (ilf.hasNext()) {
            cmd.addArgument("-C");
            cmd.addArgument(od);
            cmd.addArgument(((String) ilf.next()).substring(is));
        }
        trace("Running '" + cmd.toString() + "'");
        boolean exitCmd = cmd.run();
        // Analyse the result of the jar command
        if (exitCmd) {
            trace("The generated classes are successfully added "
                  + "in the given ejb-jar file '" + jfn + "'.");
        } else {
            throw new GenICException("Failed when adding the generated classes "
                                     + "in the given ejb-jar file '" + jfn + "'.");
        }
    }

    /**
     * Clean the intermediate generated files.
     */
    public void clean() {
        trace("Deleting " + filesToDelete.toString());
        for (Enumeration e = filesToDelete.elements(); e.hasMoreElements();) {
            String name = (String) e.nextElement();
            File f = new File(name);
            f.delete();
        }
    }

    /**
     * Display the usage
     */
    static void usage() {
        StringBuffer msg = new StringBuffer();
        msg.append("Usage: java org.objectweb.jonas_ejb.genic.GenIC -help \n");
        msg.append("         to print this help message \n");
        msg.append(" or    java org.objectweb.jonas_ejb.genic.GenIC <Options> <Input_File> \n");
        msg.append("         to generate the container-specific classes for given EJB(s). \n");
        msg.append(" \n");
        msg.append("Options include: \n");
        msg.append("       -d <output_dir>  specify where to place the generated files \n");
        msg.append("       -noaddinjar      do not add the generated classes in the given \n");
        msg.append("                        ejb-jar file \n");
        msg.append("       -nocompil        do not compile the generated source files via \n");
        msg.append("                        the java and rmi compilers \n");
        msg.append("       -novalidation    parse the XML deployment descriptors without \n");
        msg.append("                        validation \n");
        msg.append("       -javac     <opt> specify the java compiler to use \n");
        msg.append("       -javacopts <opt> specify the options to pass to the java compiler \n");
        msg.append("       -rmicopts  <opt> specify the options to pass to the rmi compiler \n");
        msg.append("       -secpropag       modify the RMI Skel. and Stub. to implement \n");
        msg.append("                        the implicit propagation of the security context \n");
        msg.append("       -mappernames     list of JORM mapper names, separated by comma \n");
        msg.append("                        (default is rdb) \n");
        msg.append("       -protocols       list of protocol, separated by comma \n");
        msg.append("                        (default is jrmp,jeremie) \n");
        msg.append("       -keepgenerated   do not delete intermediate generated source files \n");
        msg.append("       -verbose \n");
        msg.append("       -debug \n");
        msg.append(" \n");
        msg.append("       Input_File       standard deployment descriptor file's name or \n");
        msg.append("                        ejb-jar file's name \n");
        GenIC.info(msg.toString());
    }

    /**
     * Display the specified message only if verbose.
     * @param msg the message to display
     */
    void trace(String msg) {
        logger.log(verbose ? BasicLevel.INFO : BasicLevel.DEBUG, msg);
    }

    /**
     * Display the specified message.
     * @param msg the message to display
     */
    static void info(String msg) {
        logger.log(BasicLevel.INFO, msg);
    }

    /**
     * Display the specified warning message.
     * @param msg the warning message to display
     */
    static void warning(String msg) {
        logger.log(BasicLevel.WARN, msg);
    }

    /**
     * Display the specified error message.
     * @param msg the error message to display
     */
    static void error(String msg) {
        logger.log(BasicLevel.ERROR, msg);
    }


    /**
     * Display the specified error message and exits with an
     * EXIT_FAILURE status.
     * @param msg the error message to display
     */
    static void fatalError(String msg) {
        System.err.println("GenIC fatal error: " + msg);
        System.exit(EXIT_FAILURE);
    }

    /**
     * Display the specified error message and exits with an
     * EXIT_FAILURE status.
     * @param e the error to display
     */
    static void fatalError(Exception e) {
        System.err.println("GenIC fatal error: ");
        printException(e, System.err);
        e.printStackTrace(System.err);
        System.exit(EXIT_FAILURE);
    }

    /**
     * Display the exception and its nested exception if exists, on the given printstream.
     * @param e the exception to display
     * @param out the PrintStream on which the exception has to displayed
     */ 
    static void printException(Exception e, PrintStream out) {
        Exception inner = null;
        if (e instanceof GenICException
            && (inner = ((GenICException) e).inner) != null) {
            out.println(e.getMessage());
            printException(inner, out);
        } else if (e instanceof PException
                   && (inner = ((PException) e).getNestedException()) != null) {
            out.println(e.getMessage());
            printException(inner, out);
        } else if (e instanceof MedorException
                   && (inner = ((MedorException) e).getNestedException()) != null) {
            out.println(e.getMessage());
            printException(inner, out);
        } else {
            e.printStackTrace();
        }
    }

    /**
     * Create a cleaned temporary directory.
     * @return the temp directory file name
     * @throws IOException if a temp directory cannot be created.
     */
    static String createTempDir() throws IOException {       
        File tmpDir = File.createTempFile("genic", null, null);    
        tmpDir.delete();
        if (!tmpDir.mkdir()) {
            throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
        }
        return tmpDir.getAbsolutePath();
    }

    /**
     * Get the list file names recursively of the given directory
     */
    static void getFilesList(String dir, ArrayList list, boolean onlyClass) {
        File df = new File(dir);
        if (df.exists() && df.isDirectory()) {
            File [] ls = df.listFiles();
            for (int i = 0; i < ls.length; i++) {
                File f = ls[i];
                String fn = f.getAbsolutePath();
                if (f.isDirectory()) {
                    getFilesList(fn, list, onlyClass);
                } else {
                    if (!onlyClass || fn.endsWith(".class")) {
                        if (!fn.endsWith(".save")) {
                            list.add(fn);
                        }
                    }
                }
            }
        }
    }
}

class ProtocolNames {
    private TreeSet protocolNames = new TreeSet();

    ProtocolNames(String pns) {
        if (pns != null && pns.length() > 0) {
            StringTokenizer st = new StringTokenizer(pns, ",", false);
            while (st.hasMoreTokens()) {
                protocolNames.add(st.nextToken().trim());
            }
        } else {
            protocolNames.add("jrmp");
            protocolNames.add("jeremie");
        }
        if (protocolNames.contains(GenIC.CMI_RMI)) {
            protocolNames.add(GenIC.RMI_JRMP);
        }
    }

    boolean isSupported(String pn) {
        return protocolNames.contains(pn);
    }
}

