/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Sebastien Chassande-Barrioz for INRIA
 * Contributor(s): Ph Durieux (Bull)
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.lib;

import java.util.Iterator;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.EjbRelationDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbRelationshipRoleDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.FieldJdbcDesc;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.lib.JormManager;
import org.objectweb.jorm.metainfo.lib.MetaInfoPrinter;
import org.objectweb.jorm.metainfo.rdb.api.RdbClassMultiMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbExternalTable;
import org.objectweb.jorm.metainfo.rdb.api.RdbGenClassMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbJoin;
import org.objectweb.jorm.metainfo.rdb.api.RdbMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbTable;
import org.objectweb.jorm.metainfo.rdb.lib.BasicRdbMappingFactory;
import org.objectweb.jorm.metainfo.rdb.lib.RdbMappingPrinter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.wrapper.printwriter.PrintStreamImpl;

/**
 * Class to hold meta-information for Jorm
 * @author Philippe Durieux
 */
public class RdbMappingBuilder {

    /**
     * Logger for tracing
     */
    private Logger logger = null;


    /**
     * The JORM meta-information manager
     */
    private JormManager manager = null;

    /**
     * JOnAS map Entity beans on relational Databases :
     * The JORM mapper is "rdb".
     * It could be a good idea to put all the mapping information in another Class
     */
    public static final String MAPPER_NAME = "rdb";

    /**
     * Jorm need a project name to store the mapping information (Jorm Meta Information)
     */
    public static final String PROJECT_NAME = "jonas";

    /**
     * MetaInfo printer (debug)
     */
    private PrintStreamImpl pstream;
    private MetaInfoPrinter mip;

    /**
     * Constructor
     * @param dd Deployment Descriptor for the ejb-jar
     * @param logger the Logger to be used for tracing.
     * @throws DeploymentDescException Cannot build Jorm Meta Information
     */
    public RdbMappingBuilder(DeploymentCmp2Desc dd, Logger logger) throws DeploymentDescException {
        this.logger = logger;
        pstream = new PrintStreamImpl(logger);
        mip = new MetaInfoPrinter();
        mip.addMappingPrinter(new RdbMappingPrinter());

        // Init a Jorm manager mapped on Rdb. Rdb is the only mapping
        // supported by JOnAS today.
        manager = new JormManager();
        manager.setLogger(logger);
        manager.init();
        manager.addMappingFactory("rdb", new BasicRdbMappingFactory());

        try {
            // Create the jorm meta information for entity beans CMP2
            for (Iterator i = dd.getBeanDescIterator(); i.hasNext(); ) {
                BeanDesc bd = (BeanDesc) i.next();
                if (bd instanceof EntityJdbcCmp2Desc) {
                    createJormEntityMI((EntityJdbcCmp2Desc) bd);
                }
            }

            // Create the jorm meta information for Relations
            for (Iterator i = dd.getEjbRelationDescIterator(); i.hasNext(); ) {
                EjbRelationDesc relation = (EjbRelationDesc) i.next();
                createJormRelationMI(relation);
            }
        } catch (PException e) {
            logger.log(BasicLevel.ERROR, "Jorm Exception raised: " + e);
            throw new DeploymentDescException(e);
        }
    }

    /**
     * @return the Jorm MetaInfo Manager
     */
    public Manager getJormMIManager() {
        return manager;
    }

    /**
     * @return the Jorm Project Name
     */
    public static String getProjectName() {
        return PROJECT_NAME;
    }

    /**
     * Creates the Jorm Meta Information for an entity bean (CMP2)
     * @param bean the Bean Descriptor
     * @throws PException Jorm Manager reported an exception
     */
    private void createJormEntityMI(EntityJdbcCmp2Desc bean) throws PException {

        String asn = bean.getJormClassName();
        logger.log(BasicLevel.DEBUG, "Create Jorm MI for the bean: " + asn);

        // --------------------------------------------------------------------
        // create persistent class and add it to the list for Jorm Compiler
        // --------------------------------------------------------------------
        Class myejb = manager.createClass(asn);
        bean.addToJormList(myejb);
        ClassProject acp = myejb.createClassProject(PROJECT_NAME);
        RdbMapping mapping = (RdbMapping) acp.createMapping(MAPPER_NAME);
        RdbClassMultiMapping classmapping = mapping.createClassMultiMapping("to-table");
        String tablename = bean.getJdbcTableName();
        RdbTable mytable = classmapping.createRdbTable(tablename);

        // --------------------------------------------------------------------
        // CMP fields
        // --------------------------------------------------------------------
        for (Iterator i = bean.getCmpFieldDescIterator(); i.hasNext();) {
            FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
            String fieldname = fd.getName();
            logger.log(BasicLevel.DEBUG, "Create PrimitiveElement for the CMP field: " + fieldname);
            PType pt = JormType.getPType(fd.getFieldType());
            PrimitiveElement pe = myejb.createPrimitiveElement(fieldname, pt);
            mytable.createPrimitiveElementMapping(pe, fd.getJdbcFieldName(), fd.getSqlType(), fd.isPrimaryKey());
        }

        // --------------------------------------------------------------------
        // Primary Key
        // --------------------------------------------------------------------
        NameDef nds = myejb.createNameDef();
        if (bean.hasSimplePkField()) {
            // PK is a single field
            nds.setFieldName(bean.getSimplePkField().getName());
        } else {
            // Define a Composite Name (name = PK Class Name) and a NameRef
            // Sometimes, the CompositeName already exists, for example if 2 beans
            // have the same PK Class.
            boolean newcn = false;
            CompositeName cn = manager.getCompositeName(bean.getJormPKClassName());
            if (cn == null) {
                cn = manager.createCompositeName(bean.getJormPKClassName());
                bean.addToJormList(cn);
                newcn = true;
            }
            // Use the CompositeName into the NameDef
            NameRef nrs = nds.createNameRef(cn);
            // fill the composite name with pk fields and link these fields to the class fields
            for (Iterator it = bean.getCmpFieldDescIterator(); it.hasNext();) {
                FieldJdbcDesc fd = (FieldJdbcDesc) it.next();
                if (fd.isPrimaryKey()) {
                    String fieldName = fd.getName();
                    if (newcn) {
                        cn.createCompositeNameField(fieldName, JormType.getPType(fd.getFieldType()));
                    }
                    nrs.addProjection(fieldName, fieldName);
                }
            }
        }
        classmapping.createIdentifierMapping(nds);

        mip.print("Jorm MI for " + asn + ": ", myejb, pstream);
        //mip.print("JORM MI Mapping for " + asn + ": ", classmapping, pstream);
    }

    /**
     * Creates the Jorm Meta Information for a relation between CMP2 entity beans
     * @param relation the Relation Descriptor
     * @throws PException Jorm Manager reported an exception
     */
    private void createJormRelationMI(EjbRelationDesc relation) throws PException {
        logger.log(BasicLevel.DEBUG, "Create Jorm MI for the relation: " + relation.getName());
        EjbRelationshipRoleDesc role1 = relation.getRelationshipRole1();
        EjbRelationshipRoleDesc role2 = relation.getRelationshipRole2();
        createJormRoleMI(role1, role2);
        createJormRoleMI(role2, role1);
    }

    /**
     * Creates the Jorm Meta Information for a relationship role (CMR field)
     * @param sRole source EjbRelationship Role Descriptor
     * @param tRole target EjbRelationship Role Descriptor
     * @throws PException Jorm Manager reported an exception
     */
    private void createJormRoleMI(EjbRelationshipRoleDesc sRole, EjbRelationshipRoleDesc tRole)
        throws PException {

        // source bean in the relation
        EntityJdbcCmp2Desc sEntity = (EntityJdbcCmp2Desc) sRole.getSourceBean();
        String sAsn = sEntity.getJormClassName();
        Class sClass = manager.getClass(sAsn);
        logger.log(BasicLevel.DEBUG, "Create Jorm MI for the relationship role: " + sRole.getName());
        ClassProject cpro = sClass.getClassProject(PROJECT_NAME);
        RdbMapping sMapping = (RdbMapping) cpro.getMapping(MAPPER_NAME);
        RdbClassMultiMapping sClassmapping = (RdbClassMultiMapping) sMapping.getClassMapping();
        RdbTable sTable = sClassmapping.getRdbTable();

        // There is always a cmr-field, sometimes added by JOnAS, because we
        // only deal with bidirectional relations.
        String sCmr = sRole.getCmrFieldName();

        // target bean in the relation.
        EntityJdbcCmp2Desc tEntity = (EntityJdbcCmp2Desc) sRole.getTargetBean();
        String tAsn = tEntity.getJormClassName();
        Class tClass = manager.getClass(tAsn);

        // add Dependencies between beans (= same cluster of beans)
        // avoids dropping tables at each loading of beans in the same cluster.
        if (tClass != sClass) {
            sClassmapping.addDependency(tClass.getFQName());
        }

        if (sRole.isTargetMultiple()) {

            // create the gen class
            GenClassRef gcr = sClass.createGenClassRef(sCmr, sRole.getCmrFieldType().getName());

            RdbGenClassMapping gcm = null;
            RdbTable tableGC = null;
            if (sRole.isSourceMultiple()) {
                logger.log(BasicLevel.DEBUG, "Create GenClassReference for the CMR field (Many-Many): " + sCmr);
                // M <-> N case
                gcm = sMapping.createGenClassMapping("embedded-target-references", gcr);
                tableGC = gcm.createRdbTable(sRole.getRelation().getJdbcTableName());
                if (sRole.isSlave()) {
                    tableGC.setReadOnly(true);
                } else {
                    logger.log(BasicLevel.DEBUG,
                            "multivalued CMR '" + sCmr + "' will do the I/O");
                }
            } else {
                logger.log(BasicLevel.DEBUG, "Create GenClassReference for the CMR field (One-Many): " + sCmr);
                // 1 <-> N case (1 side) : store CMR on N side.
                gcm = sMapping.createGenClassMapping("embedded-target-objects", gcr);
                tableGC = gcm.createRdbTable(tEntity.getJdbcTableName());
                tableGC.setReadOnly(true); // Avoid useless I/O due to the coherency management
            }

            // How the current bean references the gen class
            // = the identifier of the source class
            // no mapping, because already done for the bean pk.
            NameDef ndsg = gcr.createRefNameDef();
            sClassmapping.createReferenceMapping("embedded-target-reference", ndsg);
            if (sEntity.hasSimplePkField()) {
                // source bean has a Simple PK.
                FieldJdbcDesc fd = (FieldJdbcDesc) sEntity.getSimplePkField();
                ndsg.setFieldName(fd.getName());
            } else {
                // source bean has a Composite PK.
                NameRef cpk = ndsg.createNameRef(sClass.getNameDef("").getNameRef().getCompositeName());
                for (Iterator i = sEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                    FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                    if (fd.isPrimaryKey()) {
                        // field names in the Composite are field names of the bean.
                        cpk.addProjection(fd.getName(), fd.getName());
                    }
                }
            }

            // How the gen class is identified (the gen class name)
            // = the identifier of the source class
            String fieldsuffix;
            if (sClass == tClass) {
                fieldsuffix = (sRole.isSlave() ? "2" : "1");
            } else {
                fieldsuffix = "";
            }
            NameDef ndid = gcr.createIdNameDef();
            if (sEntity.hasSimplePkField()) {
                // source bean has a Simple PK.
                FieldJdbcDesc fd = (FieldJdbcDesc) sEntity.getSimplePkField();
                String clafn = sAsn + "_" + fd.getName() + fieldsuffix;
                ndid.setFieldName(clafn);
                ScalarField pe = gcr.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));

                String fkname = tRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                tableGC.createPrimitiveElementMapping(pe, fkname, fd.getSqlType(), false);
                logger.log(BasicLevel.DEBUG, "GC id is mapped over the field: " + clafn + " / column:" + fkname);
            } else {
                // source bean has a Composite PK.
                NameRef cpk = ndid.createNameRef(sClass.getNameDef("").getNameRef().getCompositeName());
                for (Iterator i = sEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                    FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                    if (fd.isPrimaryKey()) {
                        String clafn = sAsn + "_" + fd.getName() + fieldsuffix;
                        ScalarField sf = gcr.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));
                        cpk.addProjection(fd.getName(), sf.getName());
                        String fkname = tRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                        tableGC.createPrimitiveElementMapping(sf, fkname, fd.getSqlType(), false);
                        logger.log(BasicLevel.DEBUG, "GC id is mapped over the field: " + clafn + " / column:" + fkname);
                    }
                }
            }
            gcm.createIdentifierMapping(ndid);

            // How the referenced class is referenced from the gen class.
            // Describes an Element of the GenClass.
            // = pk of the target bean
            if (sClass == tClass) {
                fieldsuffix = (tRole.isSlave() ? "2" : "1");
            } else {
                fieldsuffix = "";
            }
            NameDef ndr2 = gcr.createClassRef(tClass).createRefNameDef();
            gcm.createReferenceMapping("embedded-target-object", ndr2);

            if (tEntity.hasSimplePkField()) {
                // target bean has a Simple PK.
                FieldJdbcDesc fd = (FieldJdbcDesc) tEntity.getSimplePkField();
                String clafn = tAsn + "_" + fd.getName() + fieldsuffix;
                ndr2.setFieldName(clafn);
                ScalarField pe = gcr.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));
                String fkname;
                if (sRole.isSourceMultiple()) {
                    // M <-> N case
                    fkname = sRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                } else {
                    // 1 <-> N case
                    fkname = fd.getJdbcFieldName();
                }
                tableGC.createPrimitiveElementMapping(pe, fkname, fd.getSqlType(), false);
                logger.log(BasicLevel.DEBUG, "GC elem is mapped over the field: " + clafn + " / column:" + fkname);
            } else {
                // target Bean has a Composite PK.
                NameRef cpk = ndr2.createNameRef(tClass.getNameDef("").getNameRef().getCompositeName());
                for (Iterator i = tEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                    FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                    if (fd.isPrimaryKey()) {
                        String clafn = tAsn + "_" + fd.getName() + fieldsuffix;
                        ScalarField sf = gcr.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));
                        cpk.addProjection(fd.getName(), sf.getName());
                        String fkname;
                        if (sRole.isSourceMultiple()) {
                            // M <-> N case
                            fkname = sRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                        } else {
                            // 1 <-> N case
                            fkname = fd.getJdbcFieldName();
                        }
                        tableGC.createPrimitiveElementMapping(sf, fkname, fd.getSqlType(), false);
                        logger.log(BasicLevel.DEBUG, "GC elem is mapped over the field: " + clafn + " / column:" + fkname);
                    }
                }
            }
        } else {
            // Monovalued CMR field
            logger.log(BasicLevel.DEBUG, "Create ClassReference for the CMR field: " + sCmr);

            ClassRef cr = sClass.createClassRef(sCmr, tClass);
            NameDef rndf = cr.createRefNameDef();
            if (sRole.hasJdbcMapping() || tClass == sClass) {
                // Monovalued CMR field
                // source bean has a JDBC mapping for its CMR field.
                // 1 <-> N case (N side), or 1 <-> 1 case (Jdbc side)
                sClassmapping.createReferenceMapping("embedded-target-reference", rndf);
                if (tEntity.hasSimplePkField()) {
                    // target bean has a Simple PK.
                    FieldJdbcDesc fd = (FieldJdbcDesc) tEntity.getSimplePkField();
                    String clafn = sCmr + "_" + fd.getName();
                    rndf.setFieldName(clafn);
                    ScalarField sf = sClass.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));

                    String cmrmappingname = sRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                    if (cmrmappingname != null) {
                        sTable.createPrimitiveElementMapping(sf, cmrmappingname, fd.getSqlType(), false);
                    } else if (tClass == sClass) {
                        cmrmappingname = tRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                        PrimitiveElementMapping pem = sClassmapping.getPrimitiveElementMapping(tRole.getCmrFieldName() + "_" + fd.getName());
                        sClassmapping.addPrimitiveElementMapping(clafn, pem);
                    }
                    logger.log(BasicLevel.DEBUG, "CMR " + sCmr + " is mapped over the field: " + clafn + " / columname: " + cmrmappingname);
                } else {
                    // target bean has a Composite PK.
                    NameRef cpk = rndf.createNameRef(tClass.getNameDef("").getNameRef().getCompositeName());
                    for (Iterator i = tEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                        FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                        if (fd.isPrimaryKey()) {
                            String clafn = sCmr + "_" + fd.getName();
                            ScalarField sf = sClass.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));
                            cpk.addProjection(fd.getName(), sf.getName());

                            String cmrmappingname = sRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                            if (cmrmappingname != null) {
                                sTable.createPrimitiveElementMapping(sf, cmrmappingname, fd.getSqlType(), false);
                            } else if (tClass == sClass) {
                                cmrmappingname = tRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                                PrimitiveElementMapping pem = sClassmapping.getPrimitiveElementMapping(tRole.getCmrFieldName() + "_" + fd.getName());
                                sClassmapping.addPrimitiveElementMapping(clafn, pem);
                            }
                            logger.log(BasicLevel.DEBUG, "CMR " + sCmr + " is mapped over the field: " + clafn + " / columname: " + cmrmappingname);
                        }
                    }
                }
            } else {
                // Monovalued CMR field
                // The JDBC mapping for the source CMR field is done on the target bean.
                // 1 <-> 1 case (non-jdbc side)
                sClassmapping.createReferenceMapping("multi-table", rndf);
                RdbExternalTable table2 = sClassmapping.createRdbExternalTable(tEntity.getJdbcTableName());
                table2.setReadOnly(true); // Avoid useless I/O due to the coherency management
                RdbJoin joinTo2 = table2.createRdbJoin(sCmr);
                logger.log(BasicLevel.DEBUG, "CMR " + sCmr + " is mapped over the external table: " + tEntity.getJdbcTableName());
                // source pk fields compose the join
                if (sEntity.hasSimplePkField()) {
                    FieldJdbcDesc spk = (FieldJdbcDesc) sEntity.getSimplePkField();
                    String cmrmappingname = tRole.getForeignKeyJdbcName(spk.getJdbcFieldName());
                    joinTo2.addJoinColumnNames(spk.getJdbcFieldName(), cmrmappingname);
                    logger.log(BasicLevel.DEBUG, "CMR " + sCmr + ": join: " + spk.getJdbcFieldName() + " = " + cmrmappingname);
                } else {
                    for (Iterator i = sEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                        FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                        if (fd.isPrimaryKey()) {
                            String cmrmappingname = tRole.getForeignKeyJdbcName(fd.getJdbcFieldName());
                            joinTo2.addJoinColumnNames(fd.getJdbcFieldName(), cmrmappingname);
                            logger.log(BasicLevel.DEBUG, "CMR " + sCmr + ": join: " + fd.getJdbcFieldName() + " = " + cmrmappingname);
                        }
                    }
                }
                // target pk fields compose the name def of reference
                if (tEntity.hasSimplePkField()) {
                    FieldJdbcDesc tpk = (FieldJdbcDesc) tEntity.getSimplePkField();
                    String clafn = sCmr + "_" + tpk.getName();
                    rndf.setFieldName(clafn);
                    ScalarField sf = sClass.createHiddenField(clafn, JormType.getPType(tpk.getFieldType()));
                    table2.createPrimitiveElementMapping(sf, tpk.getJdbcFieldName(), tpk.getSqlType(), false);
                    logger.log(BasicLevel.DEBUG, "CMR " + sCmr + ": mapped over the field : " + clafn + " column: " + tpk.getJdbcFieldName());
                } else {
                    NameRef cpk = rndf.createNameRef(tClass.getNameDef("").getNameRef().getCompositeName());
                    for (Iterator i = tEntity.getCmpFieldDescIterator(); i.hasNext(); ) {
                        FieldJdbcDesc fd = (FieldJdbcDesc) i.next();
                        if (fd.isPrimaryKey()) {
                            String clafn = sCmr + "_" + fd.getName();
                            ScalarField sf = sClass.createHiddenField(clafn, JormType.getPType(fd.getFieldType()));
                            table2.createPrimitiveElementMapping(sf, fd.getJdbcFieldName(), fd.getSqlType(), false);
                            cpk.addProjection(fd.getName(), sf.getName());
                            logger.log(BasicLevel.DEBUG, "CMR " + sCmr + ": mapped over the field : " + clafn + " column: " + fd.getJdbcFieldName());
                        }
                    }
                }
            }
        }
        //mip.print("JORM MI for " + sAsn + ": ", sClass, pstream);
        //mip.print("JORM MI Mapping for " + sAsn + ": ", sClassmapping, pstream);
    }
}
