/*
 * @(#) OTSClientTransactionInterceptor.java	1.0 02/07/15
 *
 * Copyright (C) 2002 - INRIA (www.inria.fr)
 *
 * JOTM: Java Open Transaction Manager
 *
 * This library is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * 
 *
 */
package org.objectweb.jonas_tm.ots;

// java import
import java.io.IOException;

// omg import
import org.omg.PortableInterceptor.ClientRequestInterceptor;
import org.omg.PortableInterceptor.ClientRequestInfo;
import org.omg.PortableInterceptor.ForwardRequest;
import org.omg.PortableInterceptor.ORBInitInfo; 
import org.omg.IOP.ServiceContext;
import org.omg.CORBA.LocalObject;
import org.omg.IOP.Codec;
import org.omg.CORBA.Any;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.ORB;

// jta import
import javax.transaction.xa.Xid;
import javax.rmi.PortableRemoteObject;

// jts Import
import org.omg.CosTransactions.PropagationContext;
import org.omg.CosTransactions.PropagationContextHelper;
import org.omg.CosTransactions.TransIdentity;
import org.omg.CosTransactions.otid_t;

// Jotm import 
import org.objectweb.jonas_tm.XidImpl;
import org.objectweb.jonas_tm.Coordinator;
import org.objectweb.jonas_tm.TransactionContext;
import org.objectweb.jonas_tm.InternalTransactionContext;
import org.objectweb.jonas_tm.Current;
import org.objectweb.jonas_tm.TraceTm;


/**
 * Class <code>OTSClientTransactionInterceptor</code> is a Client Interceptor for OTS Java Client 
 * of JOTM. This Interceptor translate the Standart OTS Propagation Context to a Internal JOTM 
 * Transaction context
 * 
 * @author  Guillaume Riviere (Guillaume.Riviere@inrialpes.fr)
 * @version 1.0, 13/09/2002  
 */
public class OTSClientTransactionInterceptor extends OTSInterceptor implements ClientRequestInterceptor {

    /**
     * current object
     */
    private static Current current = null;

    /**
     * interceptor name
     */
    private String interceptorName = "OTSClientTransactionInteceptor";

    /**
     * constructor 
     */
    public OTSClientTransactionInterceptor(ORBInitInfo info) {
	super(info);
    }

    /**
     * get the name of this interceptor
     * @return name
     */
    public String name() {
	return interceptorName;
    } 

    public void destroy() {
    }

    /**
     * send client transaction context with the request, if existed.
     *
     * @param ClientRequestInfo iiop client info 
     * @exception IOException if an exception occured with the ObjectOutput
     */
    public void send_request(ClientRequestInfo jri) throws ForwardRequest {
	if (current == null) current = Current.getCurrent();
	if (current != null) {
	    try {
		// get the TransactionContext and build the Corba PropagtionContext 
		PropagationContext pContext = getCorbaPropagationContext();
		if (pContext != null) {	    
		    Any pAny = create_any();
		    PropagationContextHelper.insert(pAny, pContext); 
		    // encode the PropagationContext in a service context
		    byte[] propagationContextData = codec.encode_value(pAny);
		    ServiceContext pServiceContext = new ServiceContext(TX_CTX_ID, propagationContextData);
		    jri.add_request_service_context(pServiceContext, true);
		}
	    } catch (Exception e) {
		throw new ForwardRequest();
	    }
	}
    }

    /**
     * Receive reply interception
     * @param JClientRequestInfo jri client info 
     * @exception IOException if an exception occur with the ObjectOutput
     */
    public void receive_reply(ClientRequestInfo jri) {
	if (current == null) current = Current.getCurrent();
	if (current != null) {	
	    try {
		// get the Transaction Service Context
		ServiceContext sCtx = (ServiceContext)jri.get_reply_service_context(TX_CTX_ID);
		if (sCtx != null) { //There is a Transaction Context 
		    // unmarshall the Propagation Context
		    Any pctxAny = codec.decode_value(sCtx.context_data, PropagationContextHelper.type());
		    PropagationContext pctx = PropagationContextHelper.extract(pctxAny);
		    // get the propagation context values
		    Any specific = pctx.implementation_specific_data;
		    otid_t otid = pctx.current.otid;
		    int timeout = pctx.timeout;
		    Xid xid = new XidImpl(otid.formatID, otid.bqual_length, otid.tid);
		    Coordinator coord = null;	
		    if ((specific == null) || 
			(specific.type().kind() == TCKind.tk_null)||
			(specific.type().kind() == TCKind.tk_octet)) {
			// we do nothing
		    } else {
			coord = (Coordinator)PortableRemoteObject.narrow(specific.extract_Object(), Coordinator.class);
		    }
		    
		    // build an new TransactionContext
		    TransactionContext txCtx = new InternalTransactionContext(timeout, coord, xid);		    
		    // put into the the Current object (true for client side context)
		    current.setPropagationContext(txCtx, true);
		    // associate Thread whith Tx
		    current.associateThreadTx(xid);
		}
	    } catch (org.omg.CORBA.BAD_PARAM b) {
		// else we do nothing -> no transaction context for this call
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
	
    }

    // empty method
    public void send_poll(ClientRequestInfo jri) {
    }

    public void receive_exception(ClientRequestInfo jri) throws ForwardRequest {
	// No exeption transaction propagation for the moment
    }

    public void receive_other(ClientRequestInfo jri) throws ForwardRequest {
    }

    /**
     * Returns the CORBA PropagationContext (JTS)
     */
    private PropagationContext getCorbaPropagationContext() throws ForwardRequest {
	try {
	    // get the Transaction Context (null if there is no transaction)
	    TransactionContext txCtx = current.getPropagationContext(true);
	    if (txCtx == null) {
		return null;
	    } else {
		// build a new Propagation context	    
		
		// Get Information of TransactionContext
		Xid xid = txCtx.getXid();
		int timeout = txCtx.getTimeout();
		Coordinator coord =  txCtx.getCoordinator();
		
		// Xid is coded in the otid_t
		byte[] gtrid = xid.getGlobalTransactionId();
		byte[] bqual = xid.getBranchQualifier();
		byte[] tid = new byte[gtrid.length + bqual.length];
		System.arraycopy(bqual, 0, tid, 0, bqual.length);
		System.arraycopy(gtrid, 0, tid, bqual.length, gtrid.length);
		otid_t otid = new otid_t(xid.getFormatId(), bqual.length, tid);
		
		// current holds only otid_t
		TransIdentity curr = new TransIdentity(null, null, otid);
		
		// Create the PropagationContext
		PropagationContext pctx = new PropagationContext(timeout, curr, new TransIdentity[0], null);
		
		// Set the Coordinator.
		// Stub of the Coordinator is coded in the Any field.
		Any specific = ORB.init().create_any();
		if (coord != null) {
		    specific.insert_Object((javax.rmi.CORBA.Stub) PortableRemoteObject.toStub(coord));
		    pctx.implementation_specific_data = specific;	
		} else {
		    specific.insert_octet((byte)0);
		    pctx.implementation_specific_data = specific;
		}
		return pctx;
		
	    }
	} catch (Exception e) {
	    throw new ForwardRequest();
	}
    }
}
    
